import hjson
from tokenizers import Tokenizer
from tokenizers.models import WordPiece, BPE
from tokenizers.trainers import WordPieceTrainer, BpeTrainer
from tokenizers.pre_tokenizers import Whitespace
import argparse
from tokenizers import normalizers
from tokenizers import pre_tokenizers
import sys
sys.path.append('../')
from dlk.utils.tokenizer_util import TokenizerNormalizerFactory, PreTokenizerFactory, TokenizerPostprocessorFactory


def get_trainer(trainer_name:str):
    """TODO: Docstring for get_trainer.
    :returns: TODO

    """
    trainers = {
        "wordpiece": WordPieceTrainer,
        "bpe": BpeTrainer
    }
    if trainer_name not in trainers:
        raise KeyError(f"There is not a trainer named {trainer_name}")
    return trainers[trainer_name]


def get_model(model_name:str):
    models = {
        "wordpiece": WordPiece,
        "bpe": BPE,
    }
    if model_name not in models:
        raise KeyError(f"There is not a model named {model_name}")
    return models[model_name]


def get_processor(factory, one_processor):
    """TODO: Docstring for _get_processor.

    :factory: TODO
    :one_processor: TODO
    :returns: TODO

    """
    if isinstance(one_processor, dict):
        assert len(one_processor) == 1
        process_name, process_config = list(one_processor.items())[0]
        return factory.get(process_name)(**process_config)
    else:
        assert isinstance(one_processor, str)
        return factory.get(one_processor)()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default='./train_tokenizer.hjson',
        help="The config path.",
    )
    parser.add_argument(
        "--train_files", type=list, default=None, help="train tokenizer files."
    )
    parser.add_argument(
        "--vocab_size",
        type=int,
        default=10000,
        help=( "vocab size")
    )
    parser.add_argument(
        "--unk_token",
        type=int,
        default=10000,
        help=( "vocab size")
    )
    parser.add_argument(
        "--tokenizer_type",
        type=str,
        default='wordpiece',
        choices=['wordpiece'],
        help=( "tokenizer type [wordpiece, bpe, etc.]")
    )
    parser.add_argument(
        "--special_tokens",
        type=list,
        default=["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]"],
        help=( "tokenizer special tokens")
    )
    parser.add_argument(
        "--out_tokenizer",
        type=str,
        default="tokenizer.json",
        help=( "tokenizer output")
    )
    parser.add_argument(
        "--pre_tokenizer",
        type=list,
        default=["whitespace"],
        help=( "pretoeknizer the str")
    )
    parser.add_argument(
        "--post_processor",
        type=str,
        default='',
        help=( "post processor the str")
    )
    parser.add_argument(
        "--normalizer",
        type=list,
        default=[],
        help=( "normalizer the str before toeknizer")
    )
    args = parser.parse_args()

    if args.config:
        config_json = hjson.load(open(args.config), object_pairs_hook=dict)
        for key, value in config_json.items():
           setattr(args, key, value)

    # print(args)
    Trainer = get_trainer(args.tokenizer_type)
    Model = get_model(args.tokenizer_type)
    trainer = Trainer(vocab_size=args.vocab_size, special_tokens=args.special_tokens)

    tokenizer = Tokenizer(Model(unk_token=args.unk_token))

    pretokenizer_factory = PreTokenizerFactory()
    tokenizer_postprocessor_factory = TokenizerPostprocessorFactory()
    tokenizer_normalizer_factory = TokenizerNormalizerFactory()

    if args.pre_tokenizer:
        assert isinstance(args.pre_tokenizer, list)
        pre_tokenizers_list = []
        for one_pre_tokenizer in args.pre_tokenizer:
            pre_tokenizers_list.append(get_processor(pretokenizer_factory, one_pre_tokenizer))
        tokenizer.pre_tokenizer = pre_tokenizers.Sequence(pre_tokenizers_list)
        # print(pre_tokenizers_list)

    if args.post_processor:
        assert isinstance(args.post_processor, str) or isinstance(args.post_processor, dict)
        tokenizer.post_processor = get_processor(tokenizer_postprocessor_factory, args.post_processor)

    if args.normalizer:
        assert isinstance(args.normalizer, list)
        normalizers_list = []
        for one_normalizer in args.normalizer:
            normalizers_list.append(get_processor(tokenizer_normalizer_factory, one_normalizer))
        tokenizer.normalizer = normalizers.Sequence(normalizers_list)

    tokenizer.train(args.train_files, trainer)
    tokenizer.save(args.out_tokenizer, pretty=True)



"""
Use 'Vocabulary' map the tokens to id
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('token2id')
class Token2IDConfig(BaseConfig):
    """docstring for Token2IDConfig
        {
            "_name": "token2id",
            "config": {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "data_pair": {
                        "labels": "label_ids"
                    },
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test', 'predict'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "vocab": "label_vocab", // usually provided by the "token_gather" module
                }, //3
                "predict": "train",
                "online": "train",
            }
        }
    """

    def __init__(self, stage, config: Dict):

        super(Token2IDConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.data_pair = self.config.pop('data_pair', {})
        if self.data_set and (not self.data_pair):
            raise ValueError("You must provide 'data_pair' for token2id.")
        self.vocab = self.config.get('vocab', "")
        if self.data_set and (not self.vocab):
            raise ValueError("You must provide 'vocab' for token2id.")
        self.post_check(self.config, used=[
            "data_pair",
            "data_set",
            "vocab",
            "max_token_len",
        ])


@subprocessor_register('token2id')
class Token2ID(ISubProcessor):
    """docstring for Token2ID
    """

    def __init__(self, stage: str, config: Token2IDConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'token2id' at stage {self.stage}")
            return
        self.data_pair = config.data_pair


    def process(self, data: Dict)->Dict:

        if not self.data_set:
            return data

        vocab = Vocabulary.load(data[self.config.vocab])

        def get_index_wrap(key, x):
            """TODO: Docstring for get_index_wrap.
            """
            return vocab.auto_get_index(x[key])

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do token2id on it.')
                continue
            data_set = data['data'][data_set_name]
            for key, value in self.data_pair.items():
                get_index = partial(get_index_wrap, key)
                data_set[value] = data_set.parallel_apply(get_index, axis=1)
        return data



import torch.nn as nn
from . import embedding_register, embedding_config_register
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import pickle as pkl
from dlk.utils.logger import Logger
import torch

logger = Logger.get_logger()

@embedding_config_register('static')
class StaticEmbeddingConfig(BaseModuleConfig):
    """docstring for BasicModelConfig
    {
        "config": {
            "embedding_file": "*@*", //the embedding file, must be saved as numpy array by pickle

            "embedding_dim": "*@*",
            //if the embedding_file is a dict, you should provide the dict trace to embedding
            "embedding_trace": ".", //default the file itself is the embedding
            /*embedding_trace: "embedding", //this means the <embedding = pickle.load(embedding_file)["embedding"]>*/
            /*embedding_trace: "meta.embedding", //this means the <embedding = pickle.load(embedding_file)['meta']["embedding"]>*/
            "freeze": false, // is freeze
            "padding_idx": 0, //dropout rate
            "dropout": 0, //dropout rate
            "output_map": {},
            "input_map": {}, // required_key: provide_key
        },
        "_name": "static",
    }
    """
    def __init__(self, config: Dict):
        super(StaticEmbeddingConfig, self).__init__(config)
        config = config['config']

        embedding_file = config['embedding_file']
        embedding_file = pkl.load(open(embedding_file, 'rb'))
        embedding_trace = config["embedding_trace"]
        if embedding_trace != '.':
            traces = embedding_trace.split('.')
            for trace in traces:
                embedding_file = embedding_file[trace]
        self.embedding = embedding_file
        self.freeze = config['freeze']
        self.dropout = config['dropout']
        self.padding_idx = config['padding_idx']
        self.embedding_dim = config['embedding_dim']
        self.post_check(config, used=[
            "embedding_file",
            "embedding_dim",
            "embedding_trace",
            "freeze",
            "padding_idx",
            "dropout"
        ])


@embedding_register('static')
class StaticEmbedding(SimpleModule):
    """
    from 'input_ids' generate 'embedding'
    """

    def __init__(self, config: StaticEmbeddingConfig):
        super().__init__(config)
        self._provided_keys = set() # provided by privous module, will update by the check_keys_are_provided
        self._provide_keys = {'embedding'} # provide by this module
        self._required_keys = {'input_ids'} # required by this module
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        self.embedding = nn.Embedding.from_pretrained(torch.tensor(self.config.embedding, dtype=torch.float), freeze=self.config.freeze, padding_idx=self.config.padding_idx)
        assert self.embedding.weight.shape[-1] == self.config.embedding_dim

    def init_weight(self, method):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        logger.info(f'The static embedding is loaded the pretrained.')

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        inputs[self.get_output_name('embedding')] = self.dropout(self.embedding(inputs[self.get_input_name('input_ids')]))
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))

        return inputs



"""
    This script is for reconvert the json format data to bio data to use conlleval script to get the score
"""
import json
from tokenizers import Token, Tokenizer
tokenizer = Tokenizer.from_file('../../dlk/local_data/embeddings/glove_tokenizer.json')

data = json.load(open('./predict.json', 'r'))


def align(sentence, labels):
    """TODO: Docstring for aligh.
    :sentence: TODO
    :labels: TODO
    :returns: TODO
    """
    # predicts = line['predict_entities_info']
    encode = tokenizer.encode(sentence)
    offsets = encode.offsets
    tokens = encode.tokens
    num_label = len(labels)
    token_num = len(offsets)
    output = []
    cur_token = 0
    cur_label = 0
    while cur_token<token_num:
        if cur_label >= num_label:
            output.append((tokens[cur_token], 'O'))
            cur_token += 1
            continue
        label_start = labels[cur_label]['start']
        label_end = labels[cur_label]['end']
        label = labels[cur_label]['labels'][0]

        token_start = offsets[cur_token][0]
        token_end = offsets[cur_token][1]

        if token_start==label_start:
            output.append((tokens[cur_token], 'B-'+label))
            cur_token += 1
        elif token_start<label_start:
            output.append((tokens[cur_token], 'O'))
            cur_token += 1
        elif token_start<label_end:
            output.append((tokens[cur_token], 'I-'+label))
            cur_token += 1
        elif token_start>=label_end:
            cur_label += 1
        else:
            raise PermissionError
    # # for truth in truthes:
    return output

write = []
for line in data:
    sentence = line['sentence']
    truth_labels = line['entities_info']
    predict_labels = line['predict_entities_info']
    truth = align(sentence, truth_labels)
    predict = align(sentence, predict_labels)
    one = []
    for t, p in zip(truth, predict):
        assert t[0] == p[0]
        one.append([t[0], t[1], p[1]])
    write.append(one)

with open('predict.txt', 'w') as f:
    for line in write:
        for token in line:
            f.write(' '.join(token)+'\n')
        f.write('\n')



from typing import Dict
import torch.nn as nn
from . import loss_register, loss_config_register
from packaging import version
import torch.nn as nn
from dlk.utils.logger import Logger
from dlk.core.base_module import BaseModuleConfig
import torch

logger = Logger.get_logger()

@loss_config_register("cross_entropy")
class CrossEntropyLossConfig(BaseModuleConfig):
    """docstring for CrossEntropyLossConfig
    {
        "config": {
            "ignore_index": -1,
            "weight": null, # or a list of value for every class
            "label_smoothing": 0.0, # torch>=1.10
            "pred_truth_pair": [], # len(.) == 2, the 1st is the pred_name, 2nd is truth_name in __call__ inputs
            "schedule": [1],
            "scale": [1], # scale the loss for every schedule stage
            // "schdeule": [0.3, 1.0], # can be a list or str
            // "scale": "[0.5, 1]",
        },
        "_name": "cross_entropy",
    }
    """
    def __init__(self, config: Dict):
        super(CrossEntropyLossConfig, self).__init__(config)
        config = config['config']

        self.scale = config['scale']
        self.schedule = config['schedule']

        if isinstance(self.scale, str):
            self.scale = eval(self.scale)
        if isinstance(self.schedule, str):
            self.schedule = eval(self.schedule)

        if not isinstance(self.scale, list):
            assert isinstance(float(self.scale), float)
            self.scale = [self.scale]
        if not isinstance(self.schedule, list):
            assert isinstance(float(self.schedule), float)
            self.schedule = [self.schedule]
        assert len(self.schedule) == len(self.scale)
        assert self.schedule[-1] - 1 < 0.00001

        self.weight = config['weight']
        self.ignore_index = config['ignore_index']
        self.label_smoothing = config['label_smoothing']
        self.pred_truth_pair = config['pred_truth_pair']
        if not self.pred_truth_pair:
            raise PermissionError(f"You must provide the pred_truth_pair for loss.")
        self.post_check(config, used=[
            "ignore_index",
            "weight",
            "label_smoothing",
            "pred_truth_pair",
            "schedule",
            "scale",
        ])


@loss_register("cross_entropy")
class CrossEntropyLoss(object):
    def __init__(self, config: CrossEntropyLossConfig):
        super(CrossEntropyLoss, self).__init__()
        self.config = config
        if (version.parse(torch.__version__)>=version.parse("1.10")):
            self.cross_entropy = nn.CrossEntropyLoss(
                weight=config.weight,
                ignore_index=config.ignore_index,
                label_smoothing=config.label_smoothing
            )
        else:
            if config.label_smoothing:
                logger.info("Torch version is <1.10, so ignore label_smoothing")
            self.cross_entropy = nn.CrossEntropyLoss(
                weight=config.weight,
                ignore_index=config.ignore_index,
            )

    def update_config(self, rt_config):
        """TODO: Docstring for update_config.
        :rt_config: TODO
         {
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }
        :returns: TODO

        """
        self.current_stage = 0
        self.config.schedule = [rt_config['total_steps']*i for i in self.config.schedule]


    def calc(self, result, inputs, rt_config):
        """TODO: Docstring for get_loss.
        :returns: TODO
        """
        if rt_config['current_step']>self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        pred_name, truth_name = self.config.pred_truth_pair
        pred = result[pred_name]
        target = inputs[truth_name]
        pred = pred.reshape(-1, pred.shape[-1])
        target = target.reshape(-1)
        loss = self.cross_entropy(pred, target) * scale
        return loss

    def __call__(self, result, inputs, rt_config):
        """TODO: Docstring for __call__.
        :returns: TODO
         rt_config={
             "current_step": self.global_step,
             "current_epoch": self.current_epoch,
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }),
        """
        return self.calc(result, inputs, rt_config)



import torch.nn as nn
from . import embedding_register, embedding_config_register
from dlk.core.modules import module_register, module_config_register
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import pickle as pkl
from dlk.utils.logger import Logger
import torch

logger = Logger.get_logger()

@embedding_config_register('static_char_cnn')
class StaticCharCNNEmbeddingConfig(BaseModuleConfig):
    """docstring for BasicModelConfig
    {
        "module@cnn": {
            "_base": "conv1d",
            config: {
                in_channels: -1,
                out_channels: -1,  //will update while load embedding
                kernel_sizes: [3],
            },
        },
        "config": {
            "embedding_file": "*@*", //the embedding file, must be saved as numpy array by pickle

            //if the embedding_file is a dict, you should provide the dict trace to embedding
            "embedding_trace": ".", //default the file itself is the embedding
            /*embedding_trace: "char_embedding", //this means the <embedding = pickle.load(embedding_file)["char_embedding"]>*/
            /*embedding_trace: "meta.char_embedding", //this means the <embedding = pickle.load(embedding_file)['meta']["char_embedding"]>*/
            "freeze": false, // is freeze
            "dropout": 0, //dropout rate
            "embedding_dim": 35, //dropout rate
            "kernel_sizes": [3], //dropout rate
            "padding_idx": 0,
            "output_map": {"char_embedding": "char_embedding"},
            "input_map": {"char_ids": "char_ids"},
        },
        "_link":{
            "config.embedding_dim": ["module@cnn.config.in_channels", "module@cnn.config.out_channels"],
            "config.kernel_sizes": ["module@cnn.config.kernel_sizes"],
        },
        "_name": "static_char_cnn",
    }
    """
    def __init__(self, config: Dict):
        super(StaticCharCNNEmbeddingConfig, self).__init__(config)
        self.cnn_module_name: str = config['module@cnn']['_name']
        self.cnn_config = module_config_register.get(self.cnn_module_name)(config['module@cnn'])

        config = config['config']
        embedding_file = config['embedding_file']
        embedding_file = pkl.load(open(embedding_file, 'rb'))
        embedding_trace = config["embedding_trace"]
        if embedding_trace != '.':
            traces = embedding_trace.split('.')
            for trace in traces:
                embedding_file = embedding_file[trace]
        self.embedding = embedding_file
        self.embedding_dim = config['embedding_dim']
        self.freeze = config['freeze']
        self.dropout = config['dropout']
        self.post_check(config, used=[
            "embedding_file",
            "embedding_trace",
            "freeze",
            "dropout",
            "embedding_dim",
            "kernel_sizes",
            "padding_idx",
        ])


@embedding_register('static_char_cnn')
class StaticCharCNNEmbedding(SimpleModule):
    """
    from 'input_ids' generate 'embedding'
    """

    def __init__(self, config: StaticCharCNNEmbeddingConfig):
        super().__init__(config)
        self._provided_keys = set() # provided by privous module, will update by the check_keys_are_provided
        self._provide_keys = {'char_embedding'} # provide by this module
        self._required_keys = {'char_ids'} # required by this module
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        self.embedding = nn.Embedding.from_pretrained(torch.tensor(self.config.embedding, dtype=torch.float), freeze=self.config.freeze, padding_idx=0)
        assert self.embedding.weight.shape[-1] == self.config.embedding_dim
        self.cnn = module_register.get(self.config.cnn_module_name)(self.config.cnn_config)

    def init_weight(self, method):
        """TODO: Docstring for init_weight.
        :returns: TODO

        """
        for module in self.cnn.children():
            module.apply(method)
        logger.info(f'The static embedding is loaded the pretrained.')

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        char_ids = inputs[self.get_input_name('char_ids')]
        char_mask = (char_ids == 0).bool()
        char_embedding = self.embedding(char_ids)
        bs, seq_len, token_len, emb_dim = char_embedding.shape
        char_embedding = char_embedding.view(bs*seq_len, token_len, emb_dim)
        char_embedding = char_embedding.transpose(1, 2)
        char_embedding = self.cnn(char_embedding) # bs*seq_len, emb_dim, token_len
        word_embedding = char_embedding.masked_fill_(char_mask.view(bs*seq_len, 1, token_len), -1000).max(-1)[0].view(bs, seq_len, emb_dim).contiguous()

        inputs[self.get_output_name('char_embedding')] = self.dropout(word_embedding)
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('char_embedding')]]))
        return inputs



import hjson
import os
from typing import Dict, Union, Callable, List, Any
from dlk.utils.parser import BaseConfigParser
from dlk.utils.config import ConfigTool
from dlk.data.datamodules import datamodule_register, datamodule_config_register
from dlk.managers import manager_register, manager_config_register
from dlk.core.imodels import imodel_register, imodel_config_register
import pickle as pkl
import uuid
import json
from dlk.utils.logger import Logger
import logging

logger = Logger.get_logger()


class Train(object):
    """docstring for Trainer
        {
            "_focus": {

            },
            "_link": {},
            "_search": {},
            "config": {
                "save_dir": "*@*",  # must be provided
                "data_path": "*@*",  # must be provided
            },
            "task": {
                "_name": task_name
                ...
            }
        }
    """
    def __init__(self, config: Union[str, Dict], ckpt: str=""):
        super(Train, self).__init__()
        if not isinstance(config, dict):
            config = hjson.load(open(config), object_pairs_hook=dict)

        self.ckpt = ckpt
        self.focus = config.pop('_focus', {})
        self.configs = BaseConfigParser(config).parser_with_check()
        if self.ckpt:
            assert len(self.configs) == 1, f"Reuse the checkpoint(ckpt is not none), you must provide the (only one) config which generate the checkpoint."

        self.config_names = []
        for possible_config in self.configs:
            config_name = []
            for source, to in self.focus.items():
                config_point = possible_config
                trace = source.split('.')
                for t in trace:
                    config_point = config_point[t]
                config_name.append(to+str(config_point))
            if config_name:
                self.config_names.append('_'.join(config_name))
            else:
                self.config_names.append(possible_config['root']['_name'])

        if len(self.config_names) != len(set(self.config_names)):
            for config, name in zip(self.configs, self.config_names):
                logger.info(f"{name}:\n{json.dumps(config, indent=4)}")
            raise NameError('The config_names is not unique.')

    def run(self):
        """TODO: Docstring for run.
        :returns: TODO
        """
        logger.info(f"You have {len(self.config_names)} training config(s), they all will be run.")
        for i, (config, name) in enumerate(zip(self.configs, self.config_names)):
            logger.info(f"Runing the {i}th {name}...")
            self.run_oneturn(config, name)

    def dump_config(self, config, name):
        """TODO: Docstring for dump_config.

        :config: TODO
        :returns: TODO
        """
        log_path = os.path.join(config.get('config').get('save_dir'), name)
        os.makedirs(log_path, exist_ok=True)
        json.dump({"root":config, "_focus": self.focus}, open(os.path.join(config.get('config').get('save_dir'), name, "config.json"), 'w'), ensure_ascii=False, indent=4)
        Logger.init_file_logger("log.txt", log_path)

    def run_oneturn(self, config, name):
        """TODO: Docstring for run_oneturn.
        """

        config = config['root']
        # save configure
        self.dump_config(config, name)

        # get data
        data = self.get_data(config)

        # set datamodule
        datamodule = self.get_datamodule(config, data)

        # set training manager
        manager = self.get_manager(config, name)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        # start training
        manager.fit(model=imodel, datamodule=datamodule)
        manager.test(model=imodel, datamodule=datamodule)

    def get_data(self, config):
        """TODO: Docstring for get_data.
        :returns: TODO

        """
        self.data = pkl.load(open(config['config']['data_path'], 'rb')).get('data', {})
        return self.data

    def get_datamodule(self, config, data):
        """TODO: Docstring for get_datamodule.

        :config: TODO
        :returns: TODO

        """
        DataModule, DataModuleConfig = ConfigTool.get_leaf_module(datamodule_register, datamodule_config_register, 'datamodule', config['task']['datamodule'])
        datamodule = DataModule(DataModuleConfig, data)
        return datamodule

    def get_manager(self, config, name):
        """TODO: Docstring for get_manager.

        :config: TODO
        :returns: TODO

        """
        Manager, ManagerConfig = ConfigTool.get_leaf_module(manager_register, manager_config_register, 'manager', config.get('task').get('manager'))
        manager = Manager(ManagerConfig, rt_config={"save_dir": config.get('config').get("save_dir"), "name": name})
        return manager

    def get_imodel(self, config, data):
        """TODO: Docstring for get_imodel.

        :config: TODO
        :returns: TODO

        """
        IModel, IModelConfig = ConfigTool.get_leaf_module(imodel_register, imodel_config_register, 'imodel', config.get('task').get('imodel'))
        imodel = IModel(IModelConfig)
        if self.ckpt:
            logger.info(f"reuse the checkpoint at {self.ckpt}")
            imodel.load_from_checkpoint(self.ckpt)
        if 'valid' in data:
            imodel._origin_valid_data = data['valid']

        if 'test' in data:
            imodel._origin_test_data = data['test']

        return imodel



import torch
import torch.nn as nn
from typing import Dict, List, Callable, Any, Set
from dlk.core.modules import module_register, module_config_register
from dlk.utils.logger import Logger
from dlk.utils.config import BaseConfig
import abc

logger = Logger.get_logger()


class BaseModuleConfig(BaseConfig):
    """docstring for BaseLayerConfig"""
    def __init__(self, config: Dict):
        super(BaseModuleConfig, self).__init__(config)
        self._output_map = config['config'].pop("output_map", {})
        self._input_map = config['config'].pop('input_map', {})
        self._logits_gather_config = module_config_register.get("logits_gather")(config['config'].pop("logits_gather_config", {}))


class ModuleOutputRenameMixin:
    """Just rename the output key name by config to adapt the input field of downstream module."""
    def dict_rename(self, input: Dict, output_map: Dict[str, str])->Dict:
        if isinstance(input, dict):
            output = {}
            for key, value in input.items():
                if key in output_map:
                    output[output_map[key]] = value
                else:
                    output[key] = value
            return output
        else:
            raise PermissionError("Not Defined")

    def get_real_name(self, name, name_map):
        """TODO: Docstring for get_real_name.
        :returns: TODO

        """
        if name in name_map:
            return name_map[name]
        else:
            return name

    def get_input_name(self, name):
        """TODO: Docstring for get_input_name.
        :name: TODO
        :returns: TODO
        """
        return self.get_real_name(name, self.config._input_map)

    def get_output_name(self, name):
        """TODO: Docstring for get_input_name.
        :name: TODO
        :returns: TODO
        """
        return self.get_real_name(name, self.config._output_map)

    def set_rename(self, input: Set, output_map: Dict[str, str])->Set:
        if isinstance(input, set):
            output = set()
            for key in input:
                if key in output_map:
                    output.add(output_map[key])
                else:
                    output.add(key)
            return output
        else:
            raise PermissionError("Not Defined")


class IModuleIO(metaclass=abc.ABCMeta):
    """docstring for IModuleIO"""

    @abc.abstractmethod
    def provide_keys(self)->List[str]:
        """TODO: Docstring for provide_keys.
        :returns: TODO
        """
        pass

    @abc.abstractmethod
    def check_keys_are_provided(self, provide: List[str])->bool:
        """TODO: Docstring for check_keys_are_provided.
        :returns: TODO
        """
        pass

    def check_module_chain(self, module_list: List["BaseModule"]):
        """check the interfaces of the list of modules are alignd or not.

        :List[nn.Module]: all modules in this chain
        :returns: None, if check not passed, raise a ValueError
        """
        assert len(module_list) > 1
        result = True
        for i in range(len(module_list)-1):
            result = result and module_list[i+1].check_keys_are_provided(module_list[i].provide_keys())
            if not result:
                raise ValueError(f'The module "{module_list[i+1]._name}" is required "{", ".join(module_list[i+1].provide_keys())}", \
                    but the module "{module_list[i]._name}" provide "{", ".join(module_list[i].provide_keys())}"! ')


class IModuleStep(metaclass=abc.ABCMeta):
    """docstring for ModuleStepMixin"""


    @abc.abstractmethod
    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        raise NotImplementedError

    @abc.abstractmethod
    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """training
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        raise NotImplementedError

    @abc.abstractmethod
    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        raise NotImplementedError

    def test_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return self.validation_step(inputs)

class BaseModel(nn.Module, ModuleOutputRenameMixin, IModuleIO, IModuleStep):
    """docstring for BaseModule"""

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict forward
        """
        raise NotImplementedError

class BaseModule(nn.Module, ModuleOutputRenameMixin, IModuleIO, IModuleStep):
    """docstring for BaseModule"""

    def __init__(self, config: BaseModuleConfig):
        super(BaseModule, self).__init__()
        self._logits_gather = module_register.get("logits_gather")(config._logits_gather_config)
        self._provided_keys = set()
        self._required_keys = set()
        self._provide_keys = set()
        self.config = config # for better complete, you can rewrite this in child module

    def provide_keys(self)->Set:
        """TODO: should provide_keys in model?
        """
        return self.set_rename(self._provide_keys, self.config._output_map).union(self._provided_keys)

    def check_keys_are_provided(self, provide: Set[str])->None:
        """
        """
        self._provided_keys = provide
        provide = self.set_rename(provide, self.config._input_map)
        for required_key in self._required_keys:
            if required_key not in provide:
                raise PermissionError(f"The {self.__class__.__name__} Module required '{required_key}' as input.")

    def init_weight(self, method: Callable):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        for module in self.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict forward
        """
        raise NotImplementedError


class SimpleModule(BaseModule):
    """docstring for SimpleModule, SimpleModule, all train/predict/test/validation step call the forward"""

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict forward
        """
        raise NotImplementedError

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return self(inputs)

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """training
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return self(inputs)

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return self(inputs)

    def test_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return self(inputs)



import torch
import torch.nn as nn

class Module(nn.Module):
    """docstring for Module"""
    def __init__(self):
        super(Module, self).__init__()
        self.linear = nn.Linear(3, 4)
        self.linear2 = nn.Linear(4, 3)

    def forward(self, inp, out):
        """TODO: Docstring for forward.

        :**args: TODO
        :returns: TODO

        """
        # inp = args.pop('inp')
        x = self.linear(inp)
        x = self.linear2(x)
        return x, out

a = torch.randn(2, 1, 3)

model = Module()

script = torch.jit.trace(model, (a, a))
a = torch.randn(5, 3)
# print(script(a, a))



import hjson
import os
from typing import Dict, Union, Callable, List, Any
from dlk.utils.parser import BaseConfigParser
from dlk.utils.config import ConfigTool
from dlk.data.datamodules import datamodule_register, datamodule_config_register
from dlk.managers import manager_register, manager_config_register
from dlk.core.imodels import imodel_register, imodel_config_register
import pickle as pkl
import torch
import uuid
import json
from dlk.utils.logger import Logger

logger = Logger.get_logger()


class Predict(object):
    """docstring for Trainer
        {
            "_focus": {

            },
            "_link": {},
            "_search": {},
            "config": {
                "save_dir": "*@*",  # must be provided
                "data_path": "*@*",  # must be provided
            },
            "task": {
                "_name": task_name
                ...
            }
        }
    """
    def __init__(self, config, checkpoint):
        super(Predict, self).__init__()
        if not isinstance(config, dict):
            config = hjson.load(open(config), object_pairs_hook=dict)

        self.focus = config.pop('_focus', {})
        self.config = config
        self.ckpt = torch.load(checkpoint)
        config_name = []
        for source, to in self.focus.items():
            config_point = config
            trace = source.split('.')
            for t in trace:
                config_point = config_point[t]
            config_name.append(to+str(config_point))
        if config_name:
            name_str = '_'.join(config_name)
        else:
            name_str = config['root']['_name']
        self.name_str = name_str

    def trace(self):
        """TODO: Docstring for trace.

        :imodel: TODO
        :jj: TODO
        :returns: TODO

        """
        config = self.config['root']
        name = self.name_str
        # get data
        data = self.get_data(config)

        # set datamodule
        datamodule = self.get_datamodule(config, data)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        dataloader = datamodule.train_dataloader()
        for data in dataloader:
            # script = torch.jit.trace(imodel.model, example_inputs=data, strict=False)
            script = torch.jit.trace(imodel.model, example_inputs=data, strict=False)
            # script = torch.jit.script(imodel.model, example_inputs=data, strict=False)
            print(script)
            print(script(data))
            # imodel.model(data)
            break
        logger.error('The trace method is not implement yet.')
        raise NotImplementedError

    def predict(self):
        """TODO: Docstring for run_oneturn.
        """
        config = self.config['root']
        name = self.name_str
        # get data
        data = self.get_data(config)

        # set datamodule
        datamodule = self.get_datamodule(config, data)

        # set training manager
        manager = self.get_manager(config, name)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        # start predict
        predict_result = manager.predict(model=imodel, datamodule=datamodule)
        imodel.postprocessor(stage='predict', list_batch_outputs=predict_result, origin_data=data['predict'], rt_config={})

    def get_data(self, config):
        """TODO: Docstring for get_data.
        :returns: TODO

        """
        self.data = pkl.load(open(config['config']['data_path'], 'rb')).get('data', {})
        return self.data

    def get_datamodule(self, config, data):
        """TODO: Docstring for get_datamodule.

        :config: TODO
        :returns: TODO

        """
        DataModule, DataModuleConfig = ConfigTool.get_leaf_module(datamodule_register, datamodule_config_register, 'datamodule', config['task']['datamodule'])
        datamodule = DataModule(DataModuleConfig, data)
        return datamodule

    def get_manager(self, config, name):
        """TODO: Docstring for get_manager.

        :config: TODO
        :returns: TODO

        """
        Manager, ManagerConfig = ConfigTool.get_leaf_module(manager_register, manager_config_register, 'manager', config.get('task').get('manager'))
        manager = Manager(ManagerConfig, rt_config={"save_dir": config.get('config').get("save_dir"), "name": name})
        return manager

    def get_imodel(self, config, data):
        """TODO: Docstring for get_imodel.

        :config: TODO
        :returns: TODO

        """
        IModel, IModelConfig = ConfigTool.get_leaf_module(imodel_register, imodel_config_register, 'imodel', config.get('task').get('imodel'))
        imodel = IModel(IModelConfig, checkpoint=True)
        imodel.load_state_dict(self.ckpt['state_dict'])
        imodel.eval()
        return imodel



"""
Tokenizer the single $sentence
Or tokenizer the pair $sentence_a, $sentence_b
Generator $tokens, $input_ids, $type_ids, $special_tokens_mask, $offsets, $word_ids, $overflowing, $sequence_ids
"""
from dlk.utils.tokenizer_util import PreTokenizerFactory, TokenizerPostprocessorFactory, TokenizerNormalizerFactory
from dlk.utils.config import ConfigTool, BaseConfig
from typing import Dict, Callable
import json
from dlk.utils.logger import Logger

from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from tokenizers import normalizers
from tokenizers import pre_tokenizers

import pandas as pd
from tokenizers import Tokenizer
from tokenizers.models import WordPiece

logger = Logger.get_logger()


@subprocessor_config_register('fast_tokenizer')
class FastTokenizerConfig(BaseConfig):
    """
    docstring for GeneralTokenizerConfig
    {
        "_name": "fast_tokenizer",
        "config": {
            "train": { // you can add some whitespace surround the '&'
                "data_set": {                   // for different stage, this processor will process different part of data
                    "train": ["train", "valid", 'test'],
                    "predict": ["predict"],
                    "online": ["online"]
                },
                "config_path": "*@*",
                "truncation": {     // if this is set to None or empty, will not do trunc
                    "max_length": 512,
                    "strategy": "longest_first", // Can be one of longest_first, only_first or only_second.
                },
                "normalizer": ["nfd", "lowercase", "strip_accents", "some_processor_need_config": {config}], // if don't set this, will use the default normalizer from config
                "pre_tokenizer": [{"whitespace": {}}], // if don't set this, will use the default normalizer from config
                "post_processor": "bert", // if don't set this, will use the default normalizer from config, WARNING: not support disable  the default setting( so the default tokenizer.post_tokenizer should be null and only setting in this configure)
                "output_map": { // this is the default value, you can provide other name
                    "tokens": "tokens",
                    "ids": "input_ids",
                    "attention_mask": "attention_mask",
                    "type_ids": "type_ids",
                    "special_tokens_mask": "special_tokens_mask",
                    "offsets": "offsets",
                    "word_ids": "word_ids",
                    "overflowing": "overflowing",
                    "sequence_ids": "sequence_ids",
                }, // the tokenizer output(the key) map to the value
                "input_map": {
                    "sentence": "sentence", //for sigle input, tokenizer the "sentence"
                    "sentence_a": "sentence_a", //for pair inputs, tokenize the "sentence_a" && "sentence_b"
                    "sentence_b": "sentence_b", //for pair inputs
                },
                "deliver": "tokenizer",
                "process_data": { "is_pretokenized": false},
                "data_type": "single", // single or pair, if not provide, will calc by len(process_data)
            },
            "predict": ["train", {"deliver": null}],
            "online": ["train", {"deliver": null}],
        }
    }
    """
    def __init__(self, stage, config):
        super(FastTokenizerConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.config_path = self.config.get('config_path')
        self.normalizer = self.config.get('normalizer', "default")
        self.pre_tokenizer = self.config.get('pre_tokenizer', "default")
        self.post_processor = self.config.get('post_processor', "default")
        self.truncation = self.config["truncation"]
        self.deliver = self.config['deliver']
        self.load = self.config.get('load', None)
        self.input_map = self.config['input_map']
        self.output_map = self.config.get('output_map', { # default
            "tokens": "tokens",
            "ids": "input_ids",
            "attention_mask": "attention_mask",
            "type_ids": "type_ids",
            "special_tokens_mask": "special_tokens_mask",
            "offsets": "offsets",
            "word_ids": "word_ids",
            "overflowing": "overflowing",
            "sequence_ids": "sequence_ids",
        })
        self.process_data = self.config['process_data']
        self.data_type = self.config["data_type"]
        assert self.data_type in ['single', 'pair']
        self.post_check(self.config, used=[
            "data_set",
            "config_path",
            "truncation",
            "normalizer",
            "pre_tokenizer",
            "post_processor",
            "output_map",
            "input_map",
            "deliver",
            "process_data",
            "data_type",
        ])


@subprocessor_register('fast_tokenizer')
class FastTokenizer(ISubProcessor):
    """
    """

    def __init__(self, stage: str, config: FastTokenizerConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        if not self.config.data_set:
            logger.info(f"Skip 'fast_tokenizer' at stage {self.stage}")
            return
        self.tokenizer = Tokenizer.from_file(self.config.config_path)
        pretokenizer_factory = PreTokenizerFactory()
        tokenizer_postprocessor_factory = TokenizerPostprocessorFactory()
        tokenizer_normalizer_factory = TokenizerNormalizerFactory()

        if self.config.data_type=='single':
            self._tokenize = self._single_tokenize
        elif self.config.data_type == 'pair':
            self._tokenize = self._pair_tokenize
        else:
            raise KeyError('We only support single or pair data now.')

        if not self.config.pre_tokenizer:
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence([])
        elif self.config.pre_tokenizer != "default":
            assert isinstance(config.pre_tokenizer, list)
            pre_tokenizers_list = []
            for one_pre_tokenizer in config.pre_tokenizer:
                pre_tokenizers_list.append(self._get_processor(pretokenizer_factory, one_pre_tokenizer))
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence(pre_tokenizers_list)

        if not self.config.post_processor:
            raise KeyError("The tokenizer is not support disable default tokenizers post processer. (You can delete the config manully)")
        elif self.config.post_processor != "default":
            self.tokenizer.post_processor = self._get_processor(tokenizer_postprocessor_factory, config.post_processor)

        if not self.config.normalizer:
            self.tokenizer.normalizer = normalizers.Sequence([])
        elif self.config.normalizer != "default":
            assert isinstance(config.normalizer, list)
            normalizers_list = []
            for one_normalizer in config.normalizer:
                normalizers_list.append(self._get_processor(tokenizer_normalizer_factory, one_normalizer))
            self.tokenizer.normalizer = normalizers.Sequence(normalizers_list)

        if self.config.truncation:
            self.tokenizer.enable_truncation(max_length=self.config.truncation.get('max_length'), stride=0, strategy=self.config.truncation.get('strategy', 'longest_first'))

    def _get_processor(self, factory, one_processor):
        """TODO: Docstring for _get_processor.

        :factory: TODO
        :one_processor: TODO
        :returns: TODO

        """
        if isinstance(one_processor, dict):
            assert len(one_processor) == 1
            process_name, process_config = list(one_processor.items())[0]
            return factory.get(process_name)(**process_config)
        else:
            assert isinstance(one_processor, str)
            return factory.get(one_processor)()

    def _single_tokenize(self, one_line):
        """TODO: Docstring for _single_tokenize.

        :one_line: TODO
        :returns: TODO

        """
        sentence = one_line[self.config.input_map['sentence']]
        encode = self.tokenizer.encode(sentence, **self.config.process_data)
        return encode.tokens, encode.ids, encode.attention_mask, encode.type_ids, encode.special_tokens_mask, encode.offsets, encode.word_ids, encode.overflowing, encode.sequence_ids

    def _pair_tokenize(self, one_line):
        """TODO: Docstring for _single_tokenize.

        :one_line: TODO
        :returns: TODO

        """
        sentence_a = one_line[self.config.input_map['sentence_a']]
        sentence_b = one_line[self.config.input_map['sentence_b']]
        encode = self.tokenizer.encode(sentence_a, sentence_b, **self.config.process_data)
        return encode.tokens, encode.ids, encode.attention_mask, encode.type_ids, encode.special_tokens_mask, encode.offsets, encode.word_ids, encode.overflowing, encode.sequence_ids

    def _process(self, data):
        """TODO: Docstring for _single.

        :data: TODO
        :returns: TODO

        """
        output_map = self.config.output_map
        data[[output_map['tokens'],
            output_map['ids'],
            output_map['attention_mask'],
            output_map['type_ids'],
            output_map['special_tokens_mask'],
            output_map['offsets'],
            output_map['word_ids'],
            output_map['overflowing'],
            output_map['sequence_ids']]] = data.apply(self._tokenize, axis=1, result_type='expand')
            # WARNING: Using parallel_apply in tokenizers==0.10.3 is not fast than apply
        return data

    def process(self, data: Dict)->Dict:
        if not self.config.data_set:
            return data
        for data_set_name in self.config.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip tokenize it.')
                continue
            data_set = data['data'][data_set_name]
            data_set = self._process(data_set)
            data['data'][data_set_name] = data_set
        if self.config.deliver:
            data[self.config.deliver] = self.tokenizer.to_str()
        return data



"""
Save the processed data to $base_dir/$processed
Save the meta data(like vocab, embedding, etc.) to $base_dir/$meta
"""
from dlk.utils.config import ConfigTool, BaseConfig
from dlk.utils.logger import Logger
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
import pickle as pkl
import copy
import os

logger = Logger.get_logger()

@subprocessor_config_register('save')
class SaveConfig(BaseConfig):
    """
    Config eg.
    {
        "_name": "save",
        "config":{
            "base_dir": "."
            "train":{
                "processed": "processed_data.pkl", // all data without meta
                "meta": {
                    "meta.pkl": ['label_ids', 'embedding'] //only for next time use
                }
            },
            "predict": {
                "processed": "processed_data.pkl",
            }
        }
    },
    """

    def __init__(self, stage, config):
        super(SaveConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.base_dir:str = config.get('config').get("base_dir", ".")
        self.post_check(self.config, used=[
            "processed",
            "meta",
        ])

@subprocessor_register('save')
class Save(ISubProcessor):
    """
    """

    def __init__(self, stage: str, config: SaveConfig):
        super().__init__()
        self.stage = stage
        self.config = config.config
        if not self.config:
            logger.info(f"Skip 'save' at stage {self.stage}")
            return
        self.base_dir = config.base_dir

    def save(self, data, path):
        """TODO: Docstring for load.
        """
        if not os.path.exists(self.base_dir):
            os.mkdir(self.base_dir)
        logger.info(f"Saving file to {os.path.join(self.base_dir, path)}")
        return pkl.dump(data, open(os.path.join(self.base_dir, path), 'wb'))

    def process(self, data: Dict)->Dict:
        if not self.config:
            return data
        meta_fileds = set()
        if "meta" in self.config:
            for save_path, save_fileds in self.config['meta'].items():
                assert isinstance(save_fileds, list)
                meta_data = {}
                for field in save_fileds:
                    meta_data[field] = copy.deepcopy(data[field])
                    meta_fileds.add(field)
                self.save(meta_data, save_path)
        reserve_meta_map = {}
        for filed in meta_fileds:
            reserve_meta_map[filed] = data.pop(filed)
        if "processed" in self.config:
            self.save(data, self.config['processed'])
        data.update(reserve_meta_map)
        return data



"""
This part could merged to fast_tokenizer(it will save some time), but not all process need this part(except some special dataset like conll2003), and will make the fast_tokenizer be heavy.
Token norm:
    Love -> love
    3281 -> 0000
"""
from logging import PercentStyle
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
import pandas as pd
from dlk.utils.logger import Logger
from tokenizers.models import WordLevel
from tokenizers import Tokenizer
from tokenizers.pre_tokenizers import WhitespaceSplit

logger = Logger.get_logger()

@subprocessor_config_register('token_norm')
class TokenNormConfig(BaseConfig):
    """docstring for TokenNormConfig
        {
            "_name": "token_norm",
            "config": {
                "train":{
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test', 'predict'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "zero_digits_replaced": true,
                    "lowercase": true,
                    "extend_vocab": "", //when lowercase is true, this upper_case_vocab will collection all tokens the token is not in vocab but it's lowercase is in vocab. this is only for token gather process
                    "tokenizer": "whitespace_split",  //the path to vocab(if the token in vocab skip norm it), the file is setted to one token per line
                    "data_pair": {
                        "sentence": "norm_sentence"
                    },
                },
                "predict": "train",
                "online": "train",
            }
        }
    """
    def __init__(self, stage, config: Dict):

        super(TokenNormConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.data_pair = self.config.pop('data_pair', {})
        self.zero_digits_replaced = self.config.pop('zero_digits_replaced', True)
        self.lowercase = self.config.pop('lowercase', True)
        self.tokenizer = Tokenizer.from_file(self.config['tokenizer'])
        self.unk = self.tokenizer.model.unk_token
        self.vocab = self.tokenizer.get_vocab()
        self.do_extend_vocab = self.config['extend_vocab']
        self.prefix = self.tokenizer.model.continuing_subword_prefix
        self.post_check(self.config, used=[
            "data_set",
            "zero_digits_replaced",
            "lowercase",
            "extend_vocab",
            "tokenizer",
            "data_pair",
        ])

    def tokenize(self, seq):
        """TODO: Docstring for whitespace_split.
        :returns: TODO
        """
        encode = self.tokenizer.encode(seq)
        return encode


@subprocessor_register('token_norm')
class TokenNorm(ISubProcessor):
    """docstring for TokenNorm
    """

    def __init__(self, stage: str, config: TokenNormConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'token_norm' at stage {self.stage}")
            return
        if self.config.do_extend_vocab:
            self.extend_vocab = set()
        self._zero_digits_replaced_num = 0
        self._lower_case_num = 0
        self._lower_case_zero_digits_replaced_num = 0

    def token_norm(self, token: str):
        """norm token, the result len(result) == len(token), exp.  12348->00000
        :token: TODO
        :returns: norm token,
        """
        if self.config.zero_digits_replaced:
            norm = ''
            digit_num = 0
            for c in token:
                if c.isdigit() or c=='.':
                    norm += '0'
                    digit_num += 1
                else:
                    norm += c
            if norm in self.config.vocab or self.config.prefix+norm in self.config.vocab:
                self._zero_digits_replaced_num += 1
                return norm
            elif self.config.do_extend_vocab and digit_num == len(token) and digit_num<20:
                self.extend_vocab.add(norm)

        if self.config.lowercase:
            norm = token.lower()
            if norm in self.config.vocab or self.config.prefix+norm in self.config.vocab:
                self._lower_case_num += 1
                if self.config.do_extend_vocab:
                    self.extend_vocab.add(token)
                else:
                    raise PermissionError
                return norm

        if self.config.lowercase and self.config.zero_digits_replaced:
            norm = ''
            for c in token.lower():
                if c.isdigit() or c=='.':
                    norm += '0'
                else:
                    norm += c
            if norm in self.config.vocab or self.config.prefix+norm in self.config.vocab:
                self._lower_case_zero_digits_replaced_num += 1
                return norm
        return None

    def seq_norm(self, key:str, one_item: pd.Series):
        """TODO: Docstring for token_norm.
        :seq: TODO
        :returns: TODO
        """
        seq = one_item[key]
        norm_seq = [c for c in seq]
        encode = self.config.tokenize(seq)
        for i, token in enumerate(encode.tokens):
            if token == self.config.unk:
                token_offset = encode.offsets[i]
                prenorm_token = seq[token_offset[0]: token_offset[1]]
                norm_token = self.token_norm(prenorm_token)
                if not norm_token:
                    continue
                assert len(norm_token) == token_offset[1] - token_offset[0], f"Prenorm '{prenorm_token}', postnorm: '{norm_token}' and {len(norm_token)}!= {token_offset[1]} - {token_offset[0]}"
                norm_seq[token_offset[0]: token_offset[1]] = norm_token
        return ''.join(norm_seq)

    def process(self, data: Dict)->Dict:
        '''
        '''

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do token_norm on it.')
                continue
            data_set = data['data'][data_set_name]

            for key, value in self.config.data_pair.items():
                _seq_norm = partial(self.seq_norm, key)
                data_set[value] = data_set.apply(_seq_norm, axis=1)
                # WARNING: if you change the apply to parallel_apply, you should change the _zero_digits_replaced_num, etc. to multiprocess safely(BTW, use parallel_apply in tokenizers==0.10.3 will make the process very slow)
                # data_set[value] = data_set.apply(_seq_norm, axis=1)
        logger.info(f"We use zero digits to replace digit token num is {self._zero_digits_replaced_num}, do lowercase token num is {self._lower_case_num}, do both num is {self._lower_case_zero_digits_replaced_num}")
        if self.config.do_extend_vocab:
            logger.info(f"We will extend {len(self.extend_vocab)} tokens and deliver to {self.config.do_extend_vocab}")
            data[self.config.do_extend_vocab] = list(self.extend_vocab)
        return data



from datasets import load_dataset
from dlk.utils.logger import Logger
import copy
import json
import hjson
from dlk.process import Processor
import uuid

logger = Logger('log.txt')

def flat(data):
    """TODO: Docstring for flat.

    :data: TODO
    :returns: TODO
    """
    sentence_as = data['sentence_a']
    sentence_bs = data['sentence_b']
    uuids = data['uuid']
    labelses = data['labels']
    print(labelses)
    return [{'sentence_a': sentence_a, 'sentence_b': sentence_b, 'labels': [labels], "uuid": uuid} for sentence_a, sentence_b, labels, uuid in zip(sentence_as, sentence_bs, labelses, uuids)]

label_map = {0: "entails", 1: 'nor', 2: 'contradicts', -1: "remove"}

data = load_dataset("snli")
data = data.map(lambda one: {"sentence_a": one["hypothesis"], "sentence_b": one["premise"], 'labels': label_map[one['label']], "uuid": str(uuid.uuid1())}, remove_columns=['hypothesis', 'label', 'premise'])
data = data.filter(lambda one: one['labels'] in {'entails', 'nor', 'contradicts'})

input = {"data": {"train": flat(data['train'].to_dict()), 'valid': flat(data['validation'].to_dict()), 'test': flat(data['test'].to_dict())}}

processor = Processor('./distil_bert/prepro.hjson')
processor.fit(input)



from torch.functional import Tensor
import torch.nn as nn
from . import manager_register, manager_config_register
from typing import Dict, List
import hjson
import pytorch_lightning as pl
from dlk.utils.config import BaseConfig, ConfigTool
from dlk.utils.get_root import get_root
import os
from pytorch_lightning.callbacks import ModelCheckpoint
from dlk.core.callbacks import callback_register, callback_config_register
from dlk.utils.parser import config_parser_register
from pytorch_lightning.loggers import TensorBoardLogger


@manager_config_register('lightning')
class LightningManagerConfig(BaseConfig):
    """docstring for LightningManagerConfig
    check
    https://pytorch-lightning.readthedocs.io/en/latest/api/pytorch_lightning.trainer.trainer.html?highlight=trainer#trainer
    for para details
    """


    def __init__(self, config):
        super(LightningManagerConfig, self).__init__(config)

        manager_config = config.get('config')
        self.callbacks = self.get_callbacks_config(config)  # this is callback config, should be initialized Callback in LightningManager
        self.logger = manager_config["logger"] # True

        self.enable_checkpointing = manager_config["enable_checkpointing"] # True  use checkpoint callbac
        self.accelerator = manager_config["accelerator"] # None
        self.default_root_dir = manager_config["default_root_dir"] # None
        self.gradient_clip_val = manager_config["gradient_clip_val"] # None
        self.gradient_clip_algorithm = manager_config["gradient_clip_algorithm"] # None TODO: ? default = 'norm', can select 'norm' or 'value
        self.num_nodes = manager_config["num_nodes"] # 1
        self.num_processes = manager_config["num_processes"] # 1
        self.devices = manager_config["devices"] # None
        self.gpus = manager_config["gpus"] # None
        self.auto_select_gpus = manager_config["auto_select_gpus"] # False
        self.tpu_cores = manager_config["tpu_cores"] # None
        self.ipus = manager_config["ipus"] # None
        self.log_gpu_memory = manager_config["log_gpu_memory"] # None
        self.enable_progress_bar = manager_config["enable_progress_bar"] # True
        self.overfit_batches = eval(manager_config["overfit_batches"])
        self.track_grad_norm = manager_config["track_grad_norm"] # -1
        self.check_val_every_n_epoch = manager_config["check_val_every_n_epoch"] # 1
        self.fast_dev_run = manager_config["fast_dev_run"] # False
        self.accumulate_grad_batches = manager_config["accumulate_grad_batches"] # 1
        self.max_epochs = manager_config["max_epochs"] # None
        self.min_epochs = manager_config["min_epochs"] # None
        self.max_steps = manager_config["max_steps"] # -1
        self.min_steps = manager_config["min_steps"] # None
        self.max_time = manager_config["max_time"] # None

        self.limit_train_batches = eval(manager_config["limit_train_batches"])
        self.limit_val_batches = eval(manager_config["limit_val_batches"])
        self.limit_test_batches = eval(manager_config["limit_test_batches"])
        self.limit_predict_batches = eval(manager_config["limit_predict_batches"])
        self.val_check_interval = eval(manager_config["val_check_interval"])
        self.log_every_n_steps = manager_config["log_every_n_steps"] # 50
        self.strategy = manager_config["strategy"] # 'ddp' use ddp as default
        self.sync_batchnorm = manager_config["sync_batchnorm"] # False
        self.precision = manager_config["precision"] # 32
        self.enable_model_summary = manager_config["enable_model_summary"] # True
        self.weights_summary = manager_config["weights_summary"] # 'top'
        self.weights_save_path = manager_config["weights_save_path"] # None
        self.num_sanity_val_steps = manager_config["num_sanity_val_steps"] # 2
        self.resume_from_checkpoint = manager_config["resume_from_checkpoint"] # None
        self.profiler = manager_config["profiler"] # None  'simple', 'pytorch', etc
        self.benchmark = manager_config["benchmark"] # False
        self.deterministic = manager_config["deterministic"] # False
        self.reload_dataloaders_every_n_epochs = manager_config["reload_dataloaders_every_n_epochs"] # 0
        self.auto_lr_find = manager_config["auto_lr_find"] # False
        self.replace_sampler_ddp = manager_config["replace_sampler_ddp"] # True
        self.detect_anomaly = manager_config["detect_anomaly"] # False
        self.auto_scale_batch_size = manager_config["auto_scale_batch_size"] # False
        self.plugins = manager_config["plugins"] # None TODO: add plugins from parser plugins config
        self.amp_backend = manager_config["amp_backend"] # 'native'  pytorch>1.6
        self.amp_level = manager_config["amp_level"] # None  if not set amp_backend to "apex", don't need change this

        self.move_metrics_to_cpu = manager_config["move_metrics_to_cpu"] # False
        self.multiple_trainloader_mode = manager_config["multiple_trainloader_mode"] # 'max_size_cycle'
        self.stochastic_weight_avg = manager_config["stochastic_weight_avg"] # False
        self.terminate_on_nan = manager_config["terminate_on_nan"] # None
        self.post_check(manager_config, used=[
            "callbacks",
            "logger",
            "enable_checkpointing",
            "accelerator",
            "default_root_dir",
            "gradient_clip_val",
            "gradient_clip_algorithm",
            "num_nodes",
            "num_processes",
            "devices",
            "gpus",
            "auto_select_gpus",
            "tpu_cores",
            "ipus",
            "log_gpu_memory",
            "enable_progress_bar",
            "overfit_batches",
            "track_grad_norm",
            "check_val_every_n_epoch",
            "fast_dev_run",
            "accumulate_grad_batches",
            "max_epochs",
            "min_epochs",
            "max_steps",
            "min_steps",
            "max_time",
            "limit_train_batches",
            "limit_val_batches",
            "limit_test_batches",
            "limit_predict_batches",
            "val_check_interval",
            "log_every_n_steps",
            "strategy",
            "sync_batchnorm",
            "precision",
            "enable_model_summary",
            "weights_summary",
            "weights_save_path",
            "num_sanity_val_steps",
            "resume_from_checkpoint",
            "profiler",
            "benchmark",
            "deterministic",
            "reload_dataloaders_every_n_epochs",
            "auto_lr_find",
            "replace_sampler_ddp",
            "detect_anomaly",
            "auto_scale_batch_size",
            "plugins",
            "amp_backend",
            "amp_level",
            "move_metrics_to_cpu",
            "multiple_trainloader_mode",
            "stochastic_weight_avg",
            "terminate_on_nan",
        ])

    def get_callbacks_config(self, config):
        """parser callback and init the callbacks
        :returns: list of callbacks

        """
        callback_names = config.get("config", {}).get("callbacks", [])
        callback_configs_list = []
        for callback_name in callback_names:
            callback_config = config.get(f"callback@{callback_name}", {})
            if not callback_config:
                callback_config = hjson.load(open(os.path.join(get_root(), f'dlk/configures/core/callbacks/{callback_name}.hjson'), 'r'), object_pairs_hook=dict)
                parser_callback_config = config_parser_register.get('callback')(callback_config).parser_with_check(parser_link=False)
                assert len(parser_callback_config) == 1, f"Don't support multi callback config for one callback."
                callback_config = parser_callback_config[0]
                assert not callback_config.get("_link", {}), f"Callback don't support _link"
            callback_configs_list.append(callback_config)
        return callback_configs_list


@manager_register('lightning')
class LightningManager(object):
    """pytorch-lightning traning manager
    """

    def __init__(self, config: LightningManagerConfig, rt_config: Dict):
        """
            rt_config: {"save_dir": 'different process save to diff dir', "name": 'this config name(generally generate from _focus)'}
        """
        super().__init__()

        if config.logger:
            config.logger = TensorBoardLogger(save_dir=os.path.join(rt_config["save_dir"], rt_config["name"]), version='')
        if config.callbacks:
            config.callbacks = self.get_callbacks(config.callbacks, rt_config)
        config.__dict__.pop('_name')
        self.manager = pl.Trainer(**config.__dict__)

    def get_callbacks(self, callback_configs, rt_config):
        """parser callback and init the callbacks
        :returns: list of callbacks

        """
        callbacks_list = []

        for callback_config in callback_configs:
            Callback, CallbackConfig = ConfigTool.get_leaf_module(callback_register, callback_config_register, "callback", callback_config)
            callbacks_list.append(Callback(CallbackConfig)(rt_config=rt_config))
        return callbacks_list

    def fit(self, **inputs):
        return self.manager.fit(**inputs)

    def predict(self, **inputs):
        return self.manager.predict(**inputs)

    def test(self, **inputs):
        return self.manager.test(**inputs)

    def validate(self, **inputs):
        return self.manager.validate(**inputs)



# TODO: WIP  comming soon!!
from typing import Dict
import torch.nn as nn
from . import loss_register, loss_config_register
import torch.nn as nn
from dlk.utils.config import ConfigTool


@loss_config_register("multi_loss")
class MultiLossConfig(object):
    """docstring for CrossEntropyLossConfig
    {
        "loss@the_first": {
            config: {
                "ignore_index": -1,
                "weight": null, # or a list of value for every class
                "label_smoothing": 0.0, # torch>=1.10
                "pred_truth_pair": ["logits1", "label1"], # len(.) == 2, the 1st is the pred_name, 2nd is truth_name in __call__ inputs
                "schedule": [0.3, 0.6, 1],
                "scale": [1, 0, 0.5], # scale the loss for every schedule
                // "schdeule": [0.3, 1.0],
                // "scale": [0, 1, 0.5], # scale the loss
            },
            _name: "cross_entropy",
        },
        "loss@the_second": {
            config: {
                "pred_truth_pair": ["logits2", "label2"], # len(.) == 2, the 1st is the pred_name, 2nd is truth_name in __call__ inputs
                "schdeule": [0.3, 0.6, 1],
                "scale": [0, 1, 0.5], # scale the loss for every schedule
                // "schdeule": [0.3, 1.0],
                // "scale": [0, 1, 0.5], # scale the loss
            },
            _base: "cross_entropy",  // _name or _base is all ok
        },
        config: {
            "loss_list": ['the_first', 'the_second'],
        },
        _name: "cross_entropy",
    }
    """
    def __init__(self, config: Dict):
        super(MultiLossConfig, self).__init__()
        config = config.get('config', {})



@loss_register("multi_loss")
class MultiLoss(object):
    def __init__(self, config: MultiLossConfig):
        super(MultiLoss, self).__init__()
        self.config = config

    def get_loss(self, config):
        """get encoder config and encoder module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(loss_register, loss_config_register, "loss", config)

    def calc(self, result, inputs, rt_config):
        """TODO: Docstring for get_loss.
        :returns: TODO
        """

        loss = 0
        for i, (pred, truth) in enumerate(self.config.pred_truth_pair):
            loss = loss + self.cross_entropy(result[pred], inputs[truth]) * self.config.loss_scale[i]
        return loss

    def __call__(self, result, inputs, rt_config):
        """TODO: Docstring for __call__.
        :returns: TODO
         rt_config={
             "current_step": self.global_step,
             "current_epoch": self.current_epoch,
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }),
        """
        return self.calc(result, inputs, rt_config)



"""decoders"""
import importlib
import os
from typing import Callable, Dict, Tuple, Any
from dlk.utils.register import Register

decoder_config_register = Register("Decoder config register.")
decoder_register = Register("Decoder register.")


def import_decoders(decoders_dir, namespace):
    for file in os.listdir(decoders_dir):
        path = os.path.join(decoders_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            decoder_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + decoder_name)


# automatically import any Python files in the decoders directory
decoders_dir = os.path.dirname(__file__)
import_decoders(decoders_dir, "dlk.core.layers.decoders")



# Register  all
"""processors"""

import importlib
import os
from typing import Callable, Dict, Type
from dlk.utils.config import ConfigTool
from dlk.utils.register import Register
import abc
from pandarallel import pandarallel
pandarallel.initialize(verbose=1)
# TODO: tokenizer parall and other parall is anti
os.environ["TOKENIZERS_PARALLELISM"] = "false"

subprocessor_config_register = Register("SubProcessor config register")
subprocessor_register = Register("SubProcessor register")

class ISubProcessor(metaclass=abc.ABCMeta):
    """docstring for ISubProcessor"""


    @abc.abstractmethod
    def process(self, data: Dict)->Dict:
        """
        """
        raise NotImplementedError


def import_subprocessors(processors_dir, namespace):
    for file in os.listdir(processors_dir):
        path = os.path.join(processors_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            processor_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + processor_name)


# automatically import any Python files in the models directory
subprocessors_dir = os.path.dirname(__file__)
import_subprocessors(subprocessors_dir, "dlk.data.subprocessors")



import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import StochasticWeightAveraging


@callback_config_register('weight_average')
class StochasticWeightAveragingCallbackConfig(object):
    """docstring for StochasticWeightAveragingCallbackConfig

        {   //weight_average default
            "_name": "weight_average",
            "config": {
                "swa_epoch_start": 0.8, // swa start epoch
                "swa_lrs": null,
                    //None. Use the current learning rate of the optimizer at the time the SWA procedure starts.
                    //float. Use this value for all parameter groups of the optimizer.
                    //List[float]. A list values for each parameter group of the optimizer.
                "annealing_epochs": 10,
                "annealing_strategy": 'cos',
                "device": null, // save device, null for auto detach, if the gpu is oom, you should change this to 'cpu'
            }
        }
    """
    def __init__(self, config):
        super(StochasticWeightAveragingCallbackConfig, self).__init__()
        config = config['config']
        self.swa_epoch_start = config['swa_epoch_start']
        self.swa_lrs = config["swa_lrs"]
        self.annealing_epochs = config["annealing_epochs"]
        self.annealing_strategy = config["annealing_strategy"]
        self.device = config["device"]

@callback_register('weight_average')
class StochasticWeightAveragingCallback(object):
    """
    """

    def __init__(self, config: StochasticWeightAveragingCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict):
        """TODO: Docstring for __call__.

        :rt_config: Dict: TODO
        :returns: TODO

        """
        return StochasticWeightAveraging(**self.config.__dict__)



import torch.nn as nn
from dlk.utils.config import BaseConfig
from . import initmethod_register, initmethod_config_register
from typing import Dict, List
import torch


@initmethod_config_register('range_norm')
class RangeNormInitConfig(BaseConfig):
    """
        {
            "_name": "range_norm",
            "config": {
                "range": 0.1,
            }
        }
    """
    def __init__(self, config):
        super(RangeNormInitConfig, self).__init__(config)
        self.range = config.get("range", 0.1)
        self.post_check(config['config'], used=['range'])

@initmethod_register('range_norm')
class RangeNormInit(object):
    """for transformers
    """

    def __init__(self, config: RangeNormInitConfig):
        super().__init__()
        self.range = config.range


    def __call__(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=self.range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        elif isinstance(module, nn.Conv1d):
            module.weight.data.normal_(mean=0.0, std=self.range)
        elif isinstance(module, nn.Conv2d):
            module.weight.data.normal_(mean=0.0, std=self.range)
        elif isinstance(module, nn.Conv3d):
            module.weight.data.normal_(mean=0.0, std=self.range)



from typing import Dict
import torch.nn as nn
import torch.optim as optim
from dlk.utils.config import BaseConfig
from . import optimizer_register, optimizer_config_register, BaseOptimizer


@optimizer_config_register("sgd")
class SGDOptimizerConfig(BaseConfig):
    """
    {
        "config": {
            "lr": 1e-3,
            "momentum": 0.9,
            "dampening": 0,
            "weight_decay": 0,
            "nesterov":false,
            "optimizer_special_groups":[
            // special paramater groups set to special value, if some config key-value is not set, will use the default config in  optimizer_config.
            // You should sort the config by priority(
            //     e.g. the first group is ['linear.bias', {weight_decay: 0.1}], the second is [bias, [{weight_decay: 0.2}]], then the weight_decay of "*linea.bias*" will be 0.1, and the weight_decay of others *.bias.* will be 0.2
            // ["bias & LayerNorm.bias & LayerNorm.weight", {weight_decay: 0}]
            ]
        },
        "_name": "sgd",
    }
    """
    def __init__(self, config: Dict):
        super(SGDOptimizerConfig, self).__init__(config)
        self.config = config['config']
        self.post_check(self.config, used=[
            "lr",
            "momentum",
            "dampening",
            "weight_decay",
            "nesterov",
            "optimizer_special_groups",
        ])


@optimizer_register("sgd")
class SGDOptimizer(BaseOptimizer):
    def __init__(self, model: nn.Module, config: SGDOptimizerConfig):
        super(SGDOptimizer, self).__init__()
        self.config = config.config
        self.model = model
        self.optimizer = optim.SGD

    def get_optimizer(self):
        """TODO: Docstring for get_optimizer.

        :arg1: TODO
        :returns: TODO
        """
        return self.init_optimizer(optim.SGD, self.model, self.config)



"""
Gather tokens embedding from pretrained 'embedding_file' or init embedding(xavier_uniform init, and the range clip in 'bias_clip_range')
The tokens are from 'Tokenizer'(get_vocab) or 'Vocabulary'(word2idx) object(the two must provide only one)
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from dlk.utils.logger import Logger
from tokenizers import Tokenizer
import numpy as np

logger = Logger.get_logger()

@subprocessor_config_register('token_embedding')
class TokenEmbeddingConfig(BaseConfig):
    """Config eg.
        {
            "_name": "token_embedding",
            "config": {
                "train": { // only train stage using
                    "embedding_file": "*@*",
                    "tokenizer": null, //List of columns. Every cell must be sigle token or list of tokens or set of tokens
                    "vocab": null,
                    "deliver": "token_embedding", // output Vocabulary object (the Vocabulary of labels) name.
                    "embedding_size": 200,
                    "bias_clip_range": [0.5, 0.1], // the init embedding bias weight range, if you provide two, the larger is the up bound the lower is low bound; if you provide one value, we will use it as the bias
                }
            }
        }
    """

    def __init__(self, stage, config):
        super(TokenEmbeddingConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        if not self.config:
            return
        self.embedding_file = self.config.get("embedding_file")
        self.tokenizer = self.config.get("tokenizer")
        self.bias_clip_range = self.config['bias_clip_range']
        self.vocab = self.config['vocab']
        self.deliver = self.config.get("deliver")
        self.embedding_size = self.config["embedding_size"]
        self.post_check(self.config, used=[
            "embedding_file",
            "tokenizer",
            "vocab",
            "deliver",
            "embedding_size",
            "bias_clip_range",
        ])

@subprocessor_register('token_embedding')
class TokenEmbedding(ISubProcessor):
    """
    """
    def __init__(self, stage: str, config: TokenEmbeddingConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        if not self.config.config:
            logger.info(f"Skip 'token_embedding' at stage {self.stage}")
            return
        if config.tokenizer:
            self.tokenizer = Tokenizer.from_file(config.tokenizer)
        else:
            self.tokenizer = None
        if config.embedding_file:
            self.origin_embedding = self.get_embedding(config.embedding_file, config.embedding_size)
        else:
            self.origin_embedding = {}

    def get_embedding(self, file_path, embedding_size)->Dict[str, List[float]]:
        """TODO: Docstring for get_embedding.
        """
        embedding_dict = {}
        with open(file_path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                # if the first line is statistic info, continue
                if i==0 and len(line.split())<=embedding_size:
                    continue
                sp_line = line.split()
                if len(sp_line)<=embedding_size:
                    logger.warning(f"The {i}th line len： {len(sp_line)}, token is {sp_line[0]}")
                    continue
                word = sp_line[0]
                vector = list(map(float, sp_line[-embedding_size:]))
                embedding_dict[word] = vector
        return embedding_dict

    def update_embedding(self, embedding_dict, vocab):
        """init a embedding to the embedding_dict when some token in vocab but not in embedding_dict

        :embedding_dict: Dict
        :vocab: List[str]
        :returns: updated embedding_dict
        """
        without_embedding_tokens = 0
        fuzzy_match_tokens = 0
        bias = 0.1
        if len(self.config.bias_clip_range) == 1:
            bias = self.config.bias_clip_range[0]
        else:
            assert len(self.config.bias_clip_range) == 2, "You must provide the clip range, one or two value"
            low:float = min(self.config.bias_clip_range)
            up:float = max(self.config.bias_clip_range)
            # xavier_uniform_ init method
            bias:float = np.sqrt(6.0 / (len(vocab) + self.config.embedding_size))
            if bias>up:
                bias = up
            elif bias<low:
                bias = low
            # bias = np.sqrt(3/self.config.embedding_size)
        for token in vocab:
            if token not in embedding_dict:
                if (token.lower() not in embedding_dict) and (token.upper() not in embedding_dict):
                    embedding_dict[token] = list(np.random.uniform(-bias, bias, self.config.embedding_size))
                    without_embedding_tokens += 1
                else:
                    fuzzy_match_tokens += 1
                    if token.lower() in embedding_dict:
                        embedding_dict[token] = embedding_dict[token.lower()]
                    else:
                        embedding_dict[token] = embedding_dict[token.upper()]
        logger.info(f"All tokens num is {len(vocab)}, fuzzy mathing(lower or upper match) num is {fuzzy_match_tokens}, OOV token num is {without_embedding_tokens}")
        return embedding_dict

    def process(self, data: Dict)->Dict:
        if not self.config.config:
            return data
        if self.tokenizer is not None and self.config.vocab:
            raise PermissionError(f"The tokenizer and vocab must provide one.")
        if self.tokenizer:
            token2id = self.tokenizer.get_vocab()
            id2token = {value:key for key,value in token2id.items()}
        else:
            assert self.config.vocab, f"The tokenizer and vocab must provide one."
            vocab = data[self.config.vocab]
            token2id = vocab['word2idx']
            id2token = vocab['idx2word']
        embedding_dict = self.update_embedding(self.origin_embedding, token2id)
        embedding_mat = [embedding_dict[id2token[id]] for id in range(len(id2token))]
        data[self.config.deliver] = np.array(embedding_mat)
        return data



import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import encoder_register, encoder_config_register
from dlk.core.modules import module_config_register, module_register

@encoder_config_register("linear")
class LinearConfig(BaseModuleConfig):
    """docstring for LinearConfig
    {
        "module": {
            "_base": "linear",
        },
        "config": {
            "input_size": "*@*",
            "output_size": "*@*",
            "pool": null,
            "dropout": 0.0,
            "output_map": {},
            "input_map": {}, // required_key: provide_key
        },
        "_link":{
            "config.input_size": ["module.config.input_size"],
            "config.output_size": ["module.config.output_size"],
            "config.pool": ["module.config.pool"],
        },
        "_name": "linear",
    }
    """
    def __init__(self, config: Dict):
        super(LinearConfig, self).__init__(config)
        self.linear_config = config["module"]
        self.post_check(config['config'], used=[
            "input_size",
            "output_size",
            "pool",
            "dropout",
        ])


@encoder_register("linear")
class Linear(SimpleModule):
    def __init__(self, config: LinearConfig):
        super(Linear, self).__init__(config)
        self._provide_keys = {'embedding'}
        self._required_keys = {'embedding'}
        self._provided_keys = set()

        self.config = config

        self.linear = module_register.get('linear')(module_config_register.get('linear')(config.linear_config))

    def init_weight(self, method: Callable):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        for module in self.linear.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """
        """
        inputs[self.get_output_name("embedding")] = self.linear(inputs[self.get_input_name('embedding')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))
        return inputs



embedding_path = './data/glove.6B.100d.txt'
vocab_path = './data/embedding_vocab.txt'

vocabs = set()
with open(embedding_path, 'r', encoding='utf-8') as f:
    lines = f.readlines()
    for line in lines:
        sp_line = line.split()
        if len(sp_line)<5:
            continue
        vocabs.add(sp_line[0].strip())

for i in range(1, 20):
    if '0'*i not in vocabs:
        vocabs.add('0'*i)
        print(f"{'0'*i} is added to the vocab to prevent len(number)=={i} number oov" )
with open(vocab_path, 'w', encoding='utf-8') as f:
    f.writelines("\n".join(vocabs))



import sys
import pandas as pd
# sys.path.append('../../../')

from dlk.processors import processor_register, processor_config_register
# from dlk.models import DECODER_CONFIG_REGISTRY


def test_wordpiece():
    """TODO: Docstring for test_test.
    :returns: TODO

    """
    config= {
        'data_set': {
            'train': ['train', 'dev'],
            'predict': ['predict'],
            'online': ['online']
        },
        'config_path': './tests/processors/wp_token.json',
        "normalizer": ['nfd', 'lowercase', 'strip_accents'],
        "pre_tokenizer": [{"whitespace": {}}],
        'post_processor': 'bert',
        "filed_map": {
            "tokens": "tokens",
            "ids": "ids",
            "attention_mask": "attention_mask",
            "type_ids": "type_ids",
            "special_tokens_mask": "special_tokens_mask",
            "offsets": "offsets",
        },
        # "data_type": "single",
        # "process_data": [
            # ['sentence', { "is_pretokenized": False}],
        # ],
        "data_type": "pair",
        "process_data": [
            ['sentence_a', { "is_pretokenized": False}],
            ['sentence_b', {}],
        ],
    }
    process_config = processor_config_register.get('wordpiece_tokenizer')("online", **config)
    process = processor_register.get('wordpiece_tokenizer')("online", process_config)

    data = {}
    data['data'] = {}
    data['data']['online'] = pd.DataFrame(data={"sentence_a":['Ni a', 'wo hen hao'], "sentence_b": ['1 2', '3 4']})
    process_column_data = data['data']['online'][['sentence_a', "sentence_b"]]
    # print(process_column_data)
    # print([pair for pair in process_column_data.values])
    # print(list(process_column_data.values))
    # for i in process_column_data.values:
        # print(i)
    data = process.process(data)
    for c in data['data']['online']:
        print(c)
        print(data['data']['online'][c])
# # print(data)



import jsoneditor
import hjson
import os
import argparse


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="The config path.",
    )


    args = parser.parse_args()
    if os.path.isdir(args.config):
        config_path = os.path.join(args.config, 'config.json')
    else:
        config_path = args.config
    assert os.path.exists(config_path)

    config = hjson.load(open(config_path), object_pairs_hook=dict)

    jsoneditor.editjson(config)



from collections import Counter
import pandas as pd


class Vocabulary(object):
    """generate vocab from tokens(token or Iterable tokens)
       you can dumps the object to dict and load from dict
    """

    def __init__(self, do_strip: bool=False, unknown: str='', ignore: str="", pad: str=''):
        self.word2idx = {}
        self.idx2word = {}
        self.do_strip = do_strip
        self.word_num = 0
        self.word_count = Counter() # reserved
        self.unknown = unknown
        self.ignore = ignore
        self.pad = pad
        if ignore:
            self.word2idx[ignore] = -1
            self.idx2word[-1] = ignore
            self.word_count[ignore] += int(1e10)
        if pad:
            assert self.word_num == 0, f"The pad id must be 0"
            self.word_count[pad] += int(1e10)
            self.word2idx[pad] = self.word_num
            self.idx2word[self.word_num] = pad
            self.word_num += 1
        if unknown:
            self.word_count[unknown] += int(1e10)
            self.word2idx[unknown] = self.word_num
            self.idx2word[self.word_num] = unknown
            self.word_num += 1

    def dumps(self):
        """TODO: Docstring for dump_to_dict.
        :returns: TODO

        """
        return self.__dict__

    @classmethod
    def load(cls, attr):
        """TODO: Docstring for dump_to_dict.
        :returns: TODO

        """
        vocab = cls()
        vocab.__dict__ = attr
        return vocab

    def __getitem__(self, index):
        """TODO: Docstring for __getitem__.

        :index: TODO
        :returns: TODO

        """
        try:
            return self.idx2word[int(index)]
        except:
            raise KeyError('Undefined index: {}'.format(index))

    def auto_get_index(self, data):
        """get the index of word from this vocab
        """
        if isinstance(data, str):
            return self.get_index(data)
        elif isinstance(data, list):
            return [self.auto_get_index(subdata) for subdata in data]
        else:
            raise ValueError("Don't support the type of {}".format(data))

    def get_index(self, word):
        """get the index of word from this vocab
        """
        if self.do_strip:
            word = word.strip()
        try:
            return self.word2idx[word]
        except:
            if self.unknown:
                return self.word2idx[self.unknown]
            else:
                raise KeyError('Unkown word: {}'.format(word))

    def filter_rare(self, min_freq=1, most_common=-1):
        """TODO: Docstring for filter_rare.
        :returns: TODO

        """
        self.word2idx = {}
        self.idx2word = {}
        assert min_freq == 1 or most_common==-1, "You should set the min_freq=1 or most_common=-1."
        if most_common != -1:
            for i, (token, freq) in enumerate(self.word_count.most_common(most_common)):
                self.word2idx[token] = i
                self.idx2word[i] = token
        else:
            index = 0
            for token in self.word_count:
                if self.word_count[token]>=min_freq:
                    self.word2idx[token] = index
                    self.idx2word[index] = token
                    index += 1

    def get_word(self, index):
        """get the word of index from this vocab
        """
        try:
            return self.idx2word[int(index)]
        except:
            if index == -1:
                return '[unknown]'
            raise KeyError('Undefined index: {}'.format(index))

    def add(self, word):
        r"""
        """
        if not self.word_count[word]:
            self.word2idx[word] = self.word_num
            self.idx2word[self.word_num] = word
            self.word_num += 1
        self.word_count[word] += 1
        return self

    def auto_update(self, data):
        """auto detect data type to update the vocab

        :data: str| List[str] | Set[str] | List[List[str]]
        :returns: TODO

        """
        if isinstance(data, str):
            self.add(data)
        elif isinstance(data, list) or isinstance(data, set) or isinstance(data, pd.Series):
            self.add_from_iter(data)
        else:
            raise ValueError("Don't support the type of {}".format(data))

    def __len__(self):
        return len(self.word2idx)

    def add_from_iter(self, iterator):
        r"""add a list or set of words
        """
        for word in iterator:
            if isinstance(word, list) or isinstance(word, set):
                self.add_from_iter(word)
            elif isinstance(word, str):
                self.add(word)
            else:
                raise ValueError("Don't support the type of {}".format(word))
        return self



import torch.nn as nn
from . import embedding_register, embedding_config_register
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import torch


@embedding_config_register('identity')
class IdentityEmbeddingConfig(BaseModuleConfig):
    """docstring for IdentityEmbeddingConfig
    {
        "config": {
        },
        "_name": "identity",
    }
    """
    def __init__(self, config):
        super(IdentityEmbeddingConfig, self).__init__(config)
        self.post_check(config['config'])


@embedding_register('identity')
class IdentityEmbedding(SimpleModule):
    """
    Sequence labeling model
    """

    def __init__(self, config: IdentityEmbeddingConfig):
        super().__init__(config)
        self.config = config

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        return inputs



"""basic modules"""
import importlib
import os
from dlk.utils.register import Register

module_config_register = Register("Module config register.")
module_register = Register("Module register.")


def import_modules(modules_dir, namespace):
    for file in os.listdir(modules_dir):
        path = os.path.join(modules_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            module_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + module_name)


# automatically import any Python files in the modules directory
modules_dir = os.path.dirname(__file__)
import_modules(modules_dir, "dlk.core.modules")



import os
import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
from torch.utils.data import DataLoader, dataloader, random_split, Dataset
import pandas as pd
import time
import pytorch_lightning as pl
from typing import Dict, List, Union
from torch.nn.utils.rnn import pad_sequence
import torch.optim as optim
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint

import logging
pl_logger = logging.getLogger("ligtning")
pl_logger.setLevel(logging.INFO)
# pl_logger.propagate = False
pl_logger.info("--------------------------------------------------------------------------------------------")


# loger.info(...)
# loger.debug(...)

# TODO: train dataloader shuffle=True
    # other shuffle=False

class MNISTDataModule(pl.LightningDataModule):
    def __init__(self, dataset, collate_fn):
        super().__init__()
        self.dataset = dataset
        self.collate_fn = collate_fn

    # def train_dataloader(self):
        # return DataLoader(self.mnist_train, batch_size=self.batch_size)

    # def val_dataloader(self):
        # return DataLoader(self.mnist_val, batch_size=self.batch_size)

    # def test_dataloader(self):
        # # return DataLoader(self.mnist_test, batch_size=self.batch_size)
        # self.train_dataloader = lambda self: DataLoader(self.dataset, batch_size=4, collate_fn=self.collate_fn, pin_memory=False)

    def test_dataloader(self):
        """TODO: Docstring for predict_dataloader.
        :returns: TODO

        """
        return DataLoader(self.dataset, batch_size=4, collate_fn=self.collate_fn, pin_memory=False, num_workers=4)
    def val_dataloader(self):
        """TODO: Docstring for predict_dataloader.
        :returns: TODO

        """
        return DataLoader(self.dataset, batch_size=4, collate_fn=self.collate_fn, pin_memory=False, num_workers=4)

    def predict_dataloader(self):
        """TODO: Docstring for predict_dataloader.
        :returns: TODO

        """
        return DataLoader(self.dataset, batch_size=4, collate_fn=self.collate_fn, pin_memory=False)

    def train_dataloader(self):
        """TODO: Docstring for predict_dataloader.
        :returns: TODO

        """
        return DataLoader(self.dataset, batch_size=4, collate_fn=self.collate_fn, pin_memory=False, num_workers=4)
        # return DataLoader([])


class LitAutoEncoder(pl.LightningModule):
    def __init__(self):
        super().__init__()
        self.encoder = nn.Linear(16, 64)
        self.decoder = nn.Linear(64, 7)
        self.other = nn.Linear(8, 8)
        self.all_data = {}

    def forward(self, x: Dict[str, torch.Tensor]):
        # in lightning, forward defines the prediction/inference actions
        x = x['x']
        embedding = self.encoder(x)
        return embedding

    def predict_step(self, batch, batch_idx):
        embedding = self(batch)
        return {"embedding": embedding, "index": batch["_index"]}

    def predict_step_end(self, outputs):
        """TODO: Docstring for predict_epoch_end.
        :returns: TODO

        """
        # print("outputs: ", outputs)

    def training_step(self, batch, batch_idx):
        # training_step defined the train loop.
        # It is independent of forward
        x, y = batch['x'], batch['y']
        # y = y.squeeze()
        # print(batch['_index'])
        x = x.view(x.size(0), -1)
        z = self.encoder(x)
        x_hat = self.decoder(z)
        loss = F.cross_entropy(x_hat, y)
        # Logging to TensorBoard by default
        self.log("train_loss", loss,on_step=True, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        """TODO: Docstring for test_step.

        :batch: TODO
        :batch_idx: TODO
        :returns: TODO

        """
        x, y = batch['x'], batch['y']
        # print(batch['_index'])
        x = x.view(x.size(0), -1)
        z = self.encoder(x)
        x_hat = self.decoder(z)
        loss = F.cross_entropy(x_hat, y)
        # Logging to TensorBoard by default
        self.log("validation_loss", loss)
        return {"loss": loss.unsqueeze(0), "index": batch['_index'], "x_hat": x_hat, 'y': y}

    def test_step(self, batch, batch_idx):
        """TODO: Docstring for test_step.

        :batch: TODO
        :batch_idx: TODO
        :returns: TODO

        """
        x, y = batch['x'], batch['y']
        # print(batch['_index'])
        x = x.view(x.size(0), -1)
        z = self.encoder(x)
        x_hat = self.decoder(z)
        loss = F.cross_entropy(x_hat, y)
        # Logging to TensorBoard by default
        # self.log("train_loss", loss)
        # print(f"test_step loss: {loss}")
        # if self.local_rank == 0:
            # print("start sleep")
            # time.sleep(1)
            # print("end sleep")
        return {"loss": loss.unsqueeze(0), "index": batch['_index'], "x_hat": x_hat, 'y': y}

    def validation_epoch_end(self, outputs):
        """TODO: Docstring for test_epoch_end.
        :returns: TODO

        """
        print('batch num')
        print(len(outputs))
        # self.log("result", outputs)
        gather_output = self.all_gather(outputs)
        if self.local_rank==0:
            # print(f"test_epoch_end outputs:", outputs)
            print(f"All gather :", gather_output)
            # for key in gather_output:
                # print(f"{key}\t: {gather_output[key].shape}")
        # self.log('------------------------------', rank_zero_only=True)
        # mean = torch.mean(self.all_gather(outputs))
        # print(f"Mean: {mean}")

        # # When logging only on rank 0, don't forget to add
        # # ``rank_zero_only=True`` to avoid deadlocks on synchronization.
        # if self.trainer.is_global_zero:
            # self.log("my_reduced_metric", mean, rank_zero_only=True)
        # print(self.train_data)
        # print(f"origin outputs on rank {self.local_rank}: {outputs}")
        def proc_dist_outputs(dist_outputs):
            """gather all distributed outputs to outputs which is like in a single worker.

            :dist_outputs: the inputs of pytorch_lightning *_epoch_end when using ddp
            :returns: the inputs of pytorch_lightning *_epoch_end when only run on one worker.
            """
            outputs = []
            for dist_output in dist_outputs:
                one_output = {}
                for key in dist_output:
                    try:
                        one_output[key] = torch.cat(torch.swapaxes(dist_output[key], 0, 1).unbind(), dim=0)
                    except:
                        raise KeyError(f"{key}: {dist_output[key]}")
                outputs.append(one_output)
            return outputs
        if self.trainer.world_size>1:
            dist_outputs = self.all_gather(outputs)
            if self.local_rank in [0, -1]:
                outputs = proc_dist_outputs(dist_outputs)

        if self.local_rank in [0, -1]:
            key_all_batch_map = {}
            for batch in outputs:
                for key in batch:
                    if key not in key_all_batch_map:
                        key_all_batch_map[key] = []
                    key_all_batch_map[key].append(batch[key])
            key_all_ins_map = {}
            for key in key_all_batch_map:
                key_all_ins_map[key] = torch.cat(key_all_batch_map[key], dim=0)

            # print(key_all_ins_map)
        return outputs


    def test_epoch_end(self, outputs):
        """TODO: Docstring for test_epoch_end.
        :returns: TODO

        """
        print('batch num')
        print(len(outputs))
        # self.log("result", outputs)
        gather_output = self.all_gather(outputs)
        if self.local_rank==0:
            # print(f"test_epoch_end outputs:", outputs)
            print(f"All gather :", gather_output)
            # for key in gather_output:
                # print(f"{key}\t: {gather_output[key].shape}")
        # self.log('------------------------------', rank_zero_only=True)
        # mean = torch.mean(self.all_gather(outputs))
        # print(f"Mean: {mean}")

        # # When logging only on rank 0, don't forget to add
        # # ``rank_zero_only=True`` to avoid deadlocks on synchronization.
        # if self.trainer.is_global_zero:
            # self.log("my_reduced_metric", mean, rank_zero_only=True)
        # print(self.train_data)
        # print(f"origin outputs on rank {self.local_rank}: {outputs}")
        def proc_dist_outputs(dist_outputs):
            """gather all distributed outputs to outputs which is like in a single worker.

            :dist_outputs: the inputs of pytorch_lightning *_epoch_end when using ddp
            :returns: the inputs of pytorch_lightning *_epoch_end when only run on one worker.
            """
            outputs = []
            for dist_output in dist_outputs:
                one_output = {}
                for key in dist_output:
                    try:
                        one_output[key] = torch.cat(torch.swapaxes(dist_output[key], 0, 1).unbind(), dim=0)
                    except:
                        raise KeyError(f"{key}: {dist_output[key]}")
                outputs.append(one_output)
            return outputs
        if self.trainer.world_size>1:
            dist_outputs = self.all_gather(outputs)
            if self.local_rank in [0, -1]:
                outputs = proc_dist_outputs(dist_outputs)

        if self.local_rank in [0, -1]:
            key_all_batch_map = {}
            for batch in outputs:
                for key in batch:
                    if key not in key_all_batch_map:
                        key_all_batch_map[key] = []
                    key_all_batch_map[key].append(batch[key])
            key_all_ins_map = {}
            for key in key_all_batch_map:
                key_all_ins_map[key] = torch.cat(key_all_batch_map[key], dim=0)

            # print(key_all_ins_map)
        return outputs

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        # print("configure optim")
        # for group in optimizer.param_groups:
            # # print(f"optimizer group： {group}")
        return optimizer



class CustomDataset(Dataset):
    def __init__(self, data: pd.DataFrame):
        self.data = data

        self.get = [('x', 'float'), ('y', 'int')]
        self.type_map = {'float': torch.float, 'int': torch.long, 'bool': torch.bool}

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        one_ins = {}
        for key, key_type in self.get:
            one_ins[key] = torch.tensor(self.data.iloc[idx][key], dtype=self.type_map[key_type])
        one_ins['_index'] = torch.tensor(idx,  dtype=torch.long)
        return one_ins

# class SimpleCustomBatch:
    # def __init__(self, data):
        # # print(data)
        # # self.data = data
        # print(len(data))
        # print(data[0])
        # # transposed_data = list(zip(*data))
        # # self.inp = torch.stack(transposed_data[0], 0)
        # # self.tgt = torch.stack(transposed_data[1], 0)

    # # custom memory pinning method on custom type
    # def pin_memory(self):
        # self.inp = self.inp.pin_memory()
        # self.tgt = self.tgt.pin_memory()
        # return self


def collate_wrapper(batch):
    keys = batch[0].keys()
    data_map: Dict[str, Union[List[torch.Tensor], torch.Tensor]] = {}
    for key in keys:
        data_map[key] = []
    for key in keys:
        for one_ins in batch:
            data_map[key].append(one_ins[key])
    for key in data_map:
        try:
            data_map[key] = pad_sequence(data_map[key], batch_first=True, padding_value=0)
        except:
            if data_map[key][0].size():
                raise ValueError(f"The {data_map[key]} can not be concat by pad_sequence.")
            _data = pad_sequence([i.unsqueeze(0) for i in data_map[key]], batch_first=True, padding_value=0).squeeze()
            if not _data.size():
                _data.unsqueeze_(0)
            data_map[key] = _data
            # print( pad_sequence([i.unsqueeze(0) for i in data_map[key]], batch_first=True, padding_value=0).size()).
        # torch.cat([ ], pad=0)
    return data_map

if __name__ == "__main__":
    # pl.seed_everything(seed=21, workers=False)
    # loger = logging.getLogger('lightning')

    pl.seed_everything( workers=False)
    inp = [np.random.randn(np.random.randint(16, 17)) for _ in range(8)]
    # print(inp)
    target = [np.random.randint(0, 7) for _ in range(8)]
    # print(target)

    label = ["label"+str(i) for i in target]

    data = pd.DataFrame(data={"x": inp, 'y': target, "label": label})
    dataset = CustomDataset(data)

    # train_loader = DataLoader(dataset, batch_size=2, collate_fn=collate_wrapper, pin_memory=False)

    data_module = MNISTDataModule(dataset, collate_wrapper)
    # class MNISTDataModule(pl.LightningDataModule):
        # def __init__(self, dataset, collate_fn):
    # for data in train_loader:
        # print(data)
        # break
        # pass
    # # init model
    autoencoder = LitAutoEncoder()
    # optimizer = optim.AdamW(params=[
        # { "params": [p for p in autoencoder.encoder.parameters()], "lr": 0.3},
        # { "params": autoencoder.decoder.parameters(), "lr": 0.3},
        # { "params": autoencoder.other.parameters()},
    # ], lr=0.1
    # )
    # print(optimizer)
    # print(list(autoencoder.parameters()))







    autoencoder.train_data = data_module.dataset

    # most basic trainer, uses good defaults (auto-tensorboard, checkpoints, logs, and more)
    # trainer = pl.Trainer(gpus=8) (if you have GPUs)
    # trainer = pl.Trainer(profiler="simple", max_steps=100, val_check_interval=0.5, log_every_n_steps=10)
    logger = TensorBoardLogger(save_dir="tb_logger", version="", name="", sub_dir='test_sub_dir_2')
    # trainer = pl.Trainer(max_steps=20, val_check_interval=0.5, log_every_n_steps=2, accelerator="cpu", strategy="ddp", num_processes=3, default_root_dir='root_dir', logger=logger)

    trainer = pl.Trainer(max_steps=20, val_check_interval=0.5, log_every_n_steps=2, logger=logger, default_root_dir='test_sub_dir_2', callbacks=[ModelCheckpoint(dirpath='checkpoint_dirpath', filename='checkpoint_filename')])

# # default logger used by trainer
    # # trainer = pl.Trainer(max_steps=20, val_check_interval=0.5, log_every_n_steps=2)
    # # trainer.test(model=autoencoder)
    # trainer.fit(model=autoencoder, datamodule=data_module)
    # # print(trainer.predict(model=autoencoder, dataloaders=data_module))
    # # print(trainer.world_size)
    # # trainer.predict(model=autoencoder, dataloaders=data_module)
    # # autoencoder.load_state_dict(torch.load('./test.pt'))
    # # # print('test ')
    # # autoencoder.to_torchscript("script.sc")
    trainer.validate(autoencoder, datamodule=data_module)
    # # torch.save(autoencoder.state_dict(), './test.pt')



from . import embedding_register, embedding_config_register
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import pickle as pkl
import torch.nn as nn
import torch
import numpy as np


@embedding_config_register('random')
class RandomEmbeddingConfig(BaseModuleConfig):
    """docstring for BasicModelConfig
    {
        "config": {
            "vocab_size": "*@*",
            "embedding_dim": "*@*",
            "dropout": 0, //dropout rate
            "padding_idx": 0, //dropout rate
            "output_map": {},
            "input_map": {},
        },
        "_name": "random",
    }
    """
    def __init__(self, config: Dict):
        super(RandomEmbeddingConfig, self).__init__(config)
        config = config['config']
        self.vocab_size = config['vocab_size']
        self.embedding_dim = config['embedding_dim']
        self.dropout = config['dropout']
        self.padding_idx = config['padding_idx']
        self.post_check(config, used=[
            "vocab_size",
            "embedding_dim",
            "padding_idx",
            "dropout"
        ])


@embedding_register('random')
class RandomEmbedding(SimpleModule):
    """
    from 'input_ids' generate 'embedding'
    """

    def __init__(self, config: RandomEmbeddingConfig):
        super().__init__(config)
        self._provided_keys = set() # provided by privous module, will update by the check_keys_are_provided
        self._provide_keys = {'embedding'} # provide by this module
        self._required_keys = {'input_ids'} # required by this module
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        normal =  torch.distributions.Normal(torch.tensor([0.0]), torch.tensor([2.0/self.config.embedding_dim]))
        self.embedding = nn.Embedding.from_pretrained(normal.sample((self.config.vocab_size, self.config.embedding_dim)), padding_idx=self.config.padding_idx)

    def init_weight(self, method: Callable):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        self.embedding.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        inputs[self.get_output_name('embedding')] = self.dropout(self.embedding(inputs[self.get_input_name('input_ids')]))
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))

        return inputs



"""imodels"""
import importlib
import os
from typing import Callable, Dict, Tuple, Any
from dlk.utils.register import Register
import torch

imodel_config_register = Register("IModel config register.")
imodel_register = Register("IModel register.")

class GatherOutputMixin(object):
    """gather all the small batches output to a big batch"""


    @staticmethod
    def proc_dist_outputs(dist_outputs):
        """gather all distributed outputs to outputs which is like in a single worker.

        :dist_outputs: the inputs of pytorch_lightning *_epoch_end when using ddp
        :returns: the inputs of pytorch_lightning *_epoch_end when only run on one worker.
        """
        outputs = []
        for dist_output in dist_outputs:
            one_output = {}
            for key in dist_output:
                try:
                    one_output[key] = torch.cat(torch.swapaxes(dist_output[key], 0, 1).unbind(), dim=0)
                except:
                    raise KeyError(f"{key}: {dist_output[key]}")
            outputs.append(one_output)
        return outputs

    def gather_outputs(self, outputs):
        """TODO: Docstring for gather_outputs.

        :outputs: TODO
        :returns: TODO

        """
        if self.trainer.world_size>1:
            dist_outputs = self.all_gather(outputs)
            if self.local_rank in [0, -1]:
                outputs = self.proc_dist_outputs(dist_outputs)
        return outputs

    def concat_list_of_dict_outputs(self, outputs):
        """only support all the outputs has the same dim
        :returns: TODO
        """
        key_all_batch_map = {}
        for batch in outputs:
            for key in batch:
                if key not in key_all_batch_map:
                    key_all_batch_map[key] = []
                key_all_batch_map[key].append(batch[key])
        key_all_ins_map = {}
        for key in key_all_batch_map:
            key_all_ins_map[key] = torch.cat(key_all_batch_map[key], dim=0)

        return key_all_ins_map


def import_imodels(imodels_dir, namespace):
    for file in os.listdir(imodels_dir):
        path = os.path.join(imodels_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            imodel_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + imodel_name)


# automatically import any Python files in the imodels directory
imodels_dir = os.path.dirname(__file__)
import_imodels(imodels_dir, "dlk.core.imodels")



import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import EarlyStopping


@callback_config_register('early_stop')
class EarlyStoppingCallbackConfig(object):
    """docstring for EarlyStoppingCallbackConfig
        {
            "_name": "early_stop",
            "config":{
                "monitor": "val_loss",
                "mode": "*@*", // min or max, min for the monitor is loss, max for the monitor is acc, f1, etc.
                "patience": 3,
                "min_delta": 0.0,
                "check_on_train_epoch_end": null,
                "strict": true, // if the monitor is not right, raise error
                "stopping_threshold": null, // float, if the value is good enough, stop
                "divergence_threshold": null, // float,  if the value is so bad, stop
                "verbose": true, //verbose mode print more info
            }
        }
    """
    def __init__(self, config: Dict):
        super(EarlyStoppingCallbackConfig, self).__init__()
        config = config['config']
        self.monitor = config['monitor']
        self.mode = config['mode']
        self.patience = config["patience"]
        self.min_delta = config['min_delta']
        self.strict = config['strict']
        self.verbose = config['verbose']
        self.stopping_threshold = config['stopping_threshold']
        self.divergence_threshold = config['divergence_threshold']
        self.check_on_train_epoch_end = config["check_on_train_epoch_end"]

@callback_register('early_stop')
class EarlyStoppingCallback(object):
    """
    """

    def __init__(self, config: EarlyStoppingCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict):
        """TODO: Docstring for __call__.

        :rt_config: Dict: TODO
        :returns: TODO

        """
        return EarlyStopping(**self.config.__dict__)



"""encoders"""

import importlib
import os
from dlk.utils.register import Register


encoder_config_register = Register("Encoder config register.")
encoder_register = Register("Encoder register.")

def import_encoders(encoders_dir, namespace):
    for file in os.listdir(encoders_dir):
        path = os.path.join(encoders_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            encoder_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + encoder_name)



# automatically import any Python files in the encoders directory
encoders_dir = os.path.dirname(__file__)
import_encoders(encoders_dir, "dlk.core.layers.encoders")



from typing import Dict
from dlk.utils.config import BaseConfig
from . import scheduler_register, scheduler_config_register, BaseScheduler
from torch.optim.lr_scheduler import LambdaLR
import torch.optim as optim


@scheduler_config_register("constant_warmup")
class ConstantWarmupScheduleConfig(BaseConfig):
    """
    {
        “config“: {
            "last_epoch": -1,
            "num_warmup_steps": 0,
        },
        “_name“: "constant_warmup",
    }
    """
    def __init__(self, config: Dict):
        super(ConstantWarmupScheduleConfig, self).__init__(config)
        config = config['config']
        self.last_epoch = config["last_epoch"]
        self.num_warmup_steps = config["num_warmup_steps"]
        self.post_check(config, used=[
            "last_epoch",
            "num_warmup_steps",
        ])


@scheduler_register("constant_warmup")
class ConstantWarmupSchedule(BaseScheduler):
    def __init__(self, optimizer: optim.Optimizer, config: ConstantWarmupScheduleConfig):
        super(ConstantWarmupSchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self):
        """TODO: Docstring for get_scheduler.
        :returns: TODO
        """
        num_warmup_steps = self.config.num_warmup_steps
        last_epoch = self.config.last_epoch

        def lr_lambda(current_step: int):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1.0, num_warmup_steps))
            return 1.0

        return LambdaLR(self.optimizer, lr_lambda, last_epoch=last_epoch)



import hjson
from tokenizers import Tokenizer
from typing import Dict
import argparse



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default='./tokenizer.hjson',
        help="The config path.",
    )

    parser.add_argument(
        "--tokenizer_config",
        type=str,
        default='./tokenizer.json',
        help="The tokenizer config path.",
    )

    parser.add_argument(
        "--truncation", type=int, default=0, help="truncation length"
    )

    parser.add_argument(
        "--truncation_strategy", type=str, default='longest_first', help="longest_first, only_first, only_second", choices=['longest_first', "only_first", 'only_second']
    )

    parser.add_argument(
        "--tokenize_files", type=list, default=[], help="tokenize files."
    )
    parser.add_argument(
        "--tokenize_output_files", type=list, default=[], help="tokenize output files."
    )

    args = parser.parse_args()
    if args.config:
        config_json = hjson.load(open(args.config), object_pairs_hook=dict)
        for key, value in config_json.items():
           setattr(args, key, value)

    tokenizer = Tokenizer.from_file(args.tokenizer_config)
    # print(tokenizer.padding)

    if args.truncation:
        tokenizer.enable_truncation(args.truncation, stride=0, strategy=args.truncation_strategy)
    assert len(args.tokenize_output_files) == len(args.tokenize_files)
    for inp, out in zip(args.tokenize_files, args.tokenize_output_files):
        with open(inp, 'r') as f:
            lines = f.readlines()
        tokens = tokenizer.encode_batch(lines, is_pretokenized=False)
        with open(out, 'w') as f:
            for token in tokens:
                f.write(' '.join(token.tokens)+'\n')



import torch
import torch.nn as nn
from collections import namedtuple
from typing import Dict, List
import numpy as np
import abc
import sys
sys.path.append('../../')
from dlk.utils.base_module import BaseModule

class Module(BaseModule):
    """docstring for Module"""
    def __init__(self):
        super(Module, self).__init__()
        self.linear = nn.Linear(3, 4)
        self.linear2 = nn.Linear(4, 3)
        self.config = {}
        self.config['output_map'] = {"inp":"map_inp"}
        # self._name_map = {"inp":"map_inp"}
        self.provide = {''}
        self.required = {''}

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """training
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        raise NotImplementedError

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """valid
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        raise NotImplementedError

    def provide_keys(self)->List[str]:
        """TODO: Docstring for provide_keys.
        :returns: TODO
        """
        pass
        # raise NotImplementedError

    def check_keys_are_provided(self, provide: List[str])->bool:
        """TODO: Docstring for check_key_are_provided.
        :returns: TODO
        """
        pass

    def predict_step(self, args: Dict[str, torch.Tensor]):
        """TODO: Docstring for forward.

        :**args: TODO
        :returns: TODO

        """
        inp = args.pop('inp')
        out = args.pop('out')
        if self.training:
            print('Train')
        else:
            print("eval")

        x = self.linear(inp)
        x = self.linear2(x)
        # print(x)
        # nt = namedtuple("nt", ['inp', 'out'], defaults=[None]*2)
        result = {"inp": inp, "out": out, "x": x}
        # result = {self.name_map.get(key, key): value for key, value in result.items()}
        return self.output_rename(result)
        # return nt(out=out)


a = torch.randn(1, 3)
d = {"inp":a, "out":a}

model = Module()
# model(d)
# print(d)
# script = torch.jit.trace(model, d, strict=False)
model.eval()
# print(model(d))
script = torch.jit.script(model)
# script.train()
print(script(d))
# script.eval()
# print(script(d))
# # a = torch.randn(5, 3)
# nt = namedtuple("nt", ['inp', 'out'], defaults=None)
# out = nt._make(script(d))
# print(out.inp)
# print(out.out)
# from collections import namedtuple

# # 定义一个namedtuple类型User，并包含name，sex和age属性。
# User = namedtuple('User', ['name', 'sex', 'age'], defaults=[None]*3)

# # 创建一个User对象
# user = User(name='kongxx', sex='male')

# # # 也可以通过一个list来创建一个User对象，这里注意需要使用"_make"方法
# # user = User._make(['kongxx', 'male', 21])

# print(user)
# # User(name='user1', sex='male', age=21)

# def test(a):
    # """TODO: Docstring for test.

    # :a: TODO
    # :returns: TODO

    # """
    # print(a.inp)

# nt = namedtuple("nt", ['inp', 'out'], defaults=None)
# a = "111"
# d = nt(inp=a, out=a)
# test(d)



import hjson
import pandas as pd
from typing import Union, Dict
from dlk.utils.parser import BaseConfigParser
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.config import ConfigTool
import torch

@postprocessor_config_register('identity')
class IdentityPostProcessorConfig(IPostProcessorConfig):
    """docstring for IdentityPostProcessorConfig
    config e.g.

    """

    def __init__(self, config: Dict):
        super(IdentityPostProcessorConfig, self).__init__(config)


@postprocessor_register('identity')
class IdentityPostProcessor(IPostProcessor):
    """docstring for DataSet"""
    def __init__(self, config: IdentityPostProcessorConfig):
        super(IdentityPostProcessor, self).__init__()

    def process(self, stage, outputs, origin_data)->Dict:
        """TODO: Docstring for process.

        :data: TODO
        :returns: TODO
        """
        if 'loss' in outputs:
            return {self.loss_name_map(stage): torch.sum(outputs['loss'])}
        return {}



from transformers.models.distilbert.modeling_distilbert import DistilBertModel
from transformers.models.distilbert.configuration_distilbert import DistilBertConfig
import json
import os
import torch.nn as nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from typing import Dict
from . import module_register, module_config_register
from dlk.utils.config import BaseConfig

@module_config_register("distil_bert")
class DistilBertWrapConfig(BaseConfig):
    """docstring for DistilBertWrapConfig
    {
        "config": {
            "pretrained_model_path": "*@*",
            "from_pretrain": true,
            "freeze": false,
            "dropout": 0.0,
        },

        "_name": "distil_bert",
    }
    """

    def __init__(self, config: Dict):
        super(DistilBertWrapConfig, self).__init__(config)
        self.pretrained_model_path = config['config']['pretrained_model_path']
        self.from_pretrain = config['config']['from_pretrain']
        self.freeze = config['config']['freeze']
        self.dropout = config['config']['dropout']
        if os.path.isdir(self.pretrained_model_path):
            if os.path.exists(os.path.join(self.pretrained_model_path, 'config.json')):
                self.distil_bert_config = DistilBertConfig(**json.load(open(os.path.join(self.pretrained_model_path, 'config.json'), 'r')))
            else:
                raise PermissionError(f"config.json must in the dir {self.pretrained_model_path}")
        else:
            if os.path.isfile(self.pretrained_model_path):
                try:
                    self.distil_bert_config = DistilBertConfig(**json.load(open(self.pretrained_model_path, 'r')))
                except:
                    raise PermissionError(f"You must provide the pretrained model dir or the config file path.")
        self.post_check(config['config'], used=['pretrained_model_path', 'from_pretrain', 'freeze', 'dropout'])


@module_register("distil_bert")
class DistilBertWrap(nn.Module):
    def __init__(self, config: DistilBertWrapConfig):
        super(DistilBertWrap, self).__init__()
        self.config = config

        self.distil_bert = DistilBertModel(config.distil_bert_config)
        self.dropout = nn.Dropout(float(self.config.dropout))

    def init_weight(self, method):
        """TODO: Docstring for init_weight.
        :returns: TODO

        """
        if self.config.from_pretrain:
            self.from_pretrained()
        else:
            self.distil_bert.init_weights()

    def from_pretrained(self):
        """TODO: Docstring for init.
        :pretrained_model_path: TODO
        :returns: TODO
        """
        self.distil_bert = DistilBertModel.from_pretrained(self.config.pretrained_model_path)

    def forward(self, inputs):
        """
        """
        if self.config.freeze:
            self.distil_bert.eval()
            with torch.no_grad():
                outputs = self.distil_bert(
                    input_ids = inputs.get("input_ids", None),
                    attention_mask = inputs.get("attention_mask", None),
                    head_mask = inputs.get("head_mask", None),
                    inputs_embeds = inputs.get("inputs_embeds", None),
                    output_attentions = True,
                    output_hidden_states = True,
                    return_dict = False
                )
        else:
            outputs = self.distil_bert(
                input_ids = inputs.get("input_ids", None),
                attention_mask = inputs.get("attention_mask", None),
                head_mask = inputs.get("head_mask", None),
                inputs_embeds = inputs.get("inputs_embeds", None),
                output_attentions = True,
                output_hidden_states = True,
                return_dict = False
            )
        assert len(outputs) == 3, f"Please check transformers version, the len(outputs) is 3 for version == 4.12, and this version the output logistic of distil_bert is not as the same as bert and roberta."
        sequence_output, all_hidden_states, all_self_attentions = outputs[0], outputs[1], outputs[2]
        sequence_output = self.dropout(sequence_output)
        return sequence_output, all_hidden_states, all_self_attentions



import torch.nn as nn
from dlk.utils.config import BaseConfig
from . import initmethod_register, initmethod_config_register
from typing import Dict, List
import torch


@initmethod_config_register('range_uniform')
class RangeNormInitConfig(BaseConfig):
    """
        {
            "_name": "range_uniform",
            "config": {
                "range": 0.1,
            }
        }
    """
    def __init__(self, config):
        super(RangeNormInitConfig, self).__init__(config)
        range = config.get("range", 0.1)
        if isinstance(range, list):
            assert len(range) == 2
            self.range_from = range[0]
            self.range_to = range[1]
        else:
            assert isinstance(range, float)
            self.range_from = -abs(range)
            self.range_to = abs(range)
        self.post_check(config['config'], used='range')

@initmethod_register('range_uniform')
class RangeNormInit(object):
    """for transformers
    """

    def __init__(self, config: RangeNormInitConfig):
        super().__init__()
        self.config = config

    def __call__(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            module.weight.data.uniform_(from_=self.config.range_from, to=self.config.range_to)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.uniform_(from_=self.config.range_from, to=self.config.range_to)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        elif isinstance(module, nn.Conv1d):
            module.weight.data.uniform_(from_=self.config.range_from, to=self.config.range_to)
        elif isinstance(module, nn.Conv2d):
            module.weight.data.uniform_(from_=self.config.range_from, to=self.config.range_to)
        elif isinstance(module, nn.Conv3d):
            module.weight.data.uniform_(from_=self.config.range_from, to=self.config.range_to)



from datasets import load_dataset
from dlk.utils.logger import Logger
import copy
import json
import hjson
from dlk.process import Processor
import uuid

logger = Logger('log.txt')

def flat(data):
    """TODO: Docstring for flat.

    :data: TODO
    :returns: TODO

    """
    sentences = data['sentence']
    uuids = data['uuid']
    labelses = data['labels']
    return [{'sentence': sentece, 'labels': [labels], "uuid": uuid} for sentece, labels, uuid in zip(sentences, labelses, uuids)]

label_map = {0: "neg", 1: 'pos'}

data = load_dataset("gpt3mix/sst2")
data = data.map(lambda one: {"sentence": one["text"], 'labels': label_map[one['label']], "uuid": str(uuid.uuid1())}, remove_columns=['text', 'label'])
input = {"data": {"train": flat(data['train'].to_dict()), 'valid': flat(data['validation'].to_dict()), 'test': flat(data['test'].to_dict())}}

processor = Processor('./distil_bert/prepro.hjson')
processor.fit(input)



from typing import Dict
from dlk.utils.config import BaseConfig
from . import scheduler_register, scheduler_config_register, BaseScheduler
from torch.optim.lr_scheduler import LambdaLR
from dlk.utils.logger import Logger
import torch.optim as optim
logger = Logger.get_logger()


@scheduler_config_register("rec_decay")
class RecDecayScheduleConfig(BaseConfig):
    """ the lr=lr*1/(1+decay)
    {
        "config": {
            "last_epoch": -1,
            "num_training_steps": -1,
            "decay": 0.05,
            "epoch_training_steps": -1,
        },
        "_name": "rec_decay",
    }
    """
    def __init__(self, config: Dict):
        super(RecDecayScheduleConfig, self).__init__(config)
        config = config['config']
        self.last_epoch = config["last_epoch"]
        self.epoch_training_steps = config["epoch_training_steps"]
        self.decay = config["decay"]
        self.num_training_steps = config["num_training_steps"]
        self.post_check(config, used=[
            "last_epoch",
            "num_training_steps",
            "decay",
            "epoch_training_steps",
        ])


@scheduler_register("rec_decay")
class RecDecaySchedule(BaseScheduler):
    def __init__(self, optimizer: optim.Optimizer, config: RecDecayScheduleConfig):
        super(RecDecaySchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self):
        """TODO: Docstring for get_scheduler.
        :returns: TODO
        """
        num_training_steps = self.config.num_training_steps
        epoch_training_steps = self.config.epoch_training_steps
        decay = self.config.decay
        last_epoch = self.config.last_epoch
        logger.warning(f"The calculated Total Traning Num is {num_training_steps}, the Epoch training Steps is {epoch_training_steps}. Please check it carefully.")

        def lr_lambda(current_step: int):
            cur_epoch = (current_step+1)//epoch_training_steps if epoch_training_steps!=0 else 0
            # return 1/(1+decay*cur_epoch)
            return 1/((1+decay)**cur_epoch)
        return LambdaLR(self.optimizer, lr_lambda, last_epoch)



from setuptools import setup, find_packages
import os


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

configures = package_files('dlk/configures')
with open('README.md', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read()

pkgs = [p for p in find_packages() if p.startswith('dlk')]

setup(
    name='dlk',
    version='0.0.2',
    url='https://github.com/cstsunfu/dlk',
    description='dlk: Deep Learning Kit',
    long_description=readme,
    long_description_content_type='text/markdown',
    license=license,
    author='cstsunfu',
    author_email='cstsunfu@gmail.com',
    python_requires='>=3.8',
    package_data={"": configures},
    include_package_data=True,
    packages=pkgs,
    install_requires=requirements.strip().split('\n'),
)
